"""
**MeanFunction Module**

The MeanFunction module contains classes used for constructing mean functions for GP emulators.
A base ``MeanBase`` class is provided, which implements basic operations to combine fixed
functions and fitting parameters. The basic operations ``f1 + f2``, ``f1*f2``, ``f1**f2``
and ``f1(f2)`` are available, though not all possible combinations will make sense.
Particular cases where combinations do not make sense often use classes that represent free
fitting parameters, or if you attempt to raise a mean function to a power that is not
independent of the inputs. These operations will create new derived classes ``MeanSum``,
``MeanProduct``, ``MeanPower``, and ``MeanComposite``, from which more complex
regression functions can be formed. The derived sum, product, power, and composite mean classes
call the necessary methods to compute the function and derivatives from the more basic
classes and then combine them using sum, product, power, and chain rules for function evaluation
and derivatives.

The basic building blocks are fixed mean functions, derived from ``FixedMean``, and free
parameters, represented by the ``Coefficient`` class. Incuded fixed functions include
``ConstantMean`` and ``LinearMean``. Additional derived ``FixedMean`` functions can be
created by initializing a new ``FixedMean`` instance where the user provides a fixed
function and its derivative, and these can be combined to form arbitrarily complex mean functions.
Future improvements will extend the number of pre-defined function options.

One implementation note: ``CompositeMean`` does not implement the Hessian, as computing this
requires mixed partials involving inputs and parameters that are not normally implemented.
If a composite mean is required with a Hessian Computation, the user must implement this.

Additionally, note that given mean function may have a number of parameters that depends on
the shape of the input. Since the mean function does not store input, but rather provides
a way to collate functions and derivatives together in a single object, the number of parameters
can vary based on the inputs. This is particularly true for the provided ``PolynomialMean``
class, which fits a polynomial function of a fixed degree to each input parameter. Thus,
the number of parameters depends on the input shape.

In addition to manually creating a mean function by composing fixed functions and fitting parameters,
a ``MeanBase`` subclass can be created by using the ``MeanFunction`` function. ``MeanFunction``
is a functional interface for creating ``MeanBase`` subclasses from a string formula.
The formula langauge supports the operations described above as expected (see below for some
examples), with the option to first parse the formula using the Python library ``patsy``
before converting the terms to the respective subclasses of ``MeanBase``. Formulas specify
input variables using either ``x[i]`` or ``inputs[i]`` to represent the dependent variables,
and can explicitly include a leading ``"y ="`` or ``"y ~"`` (which will be ignored). Optionally,
named variables can be mapped to input dimensions by providing a dictionary mapping strings
to integer indices. Any other variables in the formula will be assumed to be fitting
coefficients. Note that the formula parser does not make any effort to simplify expressions
(such as having identical terms or a term with redundant fitting parameters), so it is up
to the user to get things correct. Converting a mean function instance to a string can
be very helpful in determining if the parsing led to any problems, see below.

Example: ::

    >>> from mogp_emulator.MeanFunction import Coefficient, LinearMean, MeanFunction
    >>> mf1 = Coefficient() + Coefficient()*LinearMean()
    >>> print(mf1)
    c + c*x[0]
    >>> mf2 = LinearMean(1)*LinearMean(2)
    >>> print(mf2)
    x[1]*x[2]
    >>> mf3 = mf1(mf2)
    >>> print(mf3)
    c + c*x[1]*x[2]
    >>> mf4 = Coefficient()*LinearMean()**2
    >>> print(mf4)
    c*x[0]^2
    >>> mf5 = MeanFunction("x[0]")
    >>> print(mf5)
    c + c*x[0]
    >>> mf6 = MeanFunction("y = a + b*x[0]", use_patsy=False)
    >>> print(mf6)
    c + c*x[0]
    >>> mf7 = MeanFunction("a*b", {"a": 0, "b": 1})
    >>> print(mf7)
    c + c*x[0] + c*x[1] + c*x[0]*x[1]
"""
import numpy as np
from functools import partial
from inspect import signature
from mogp_emulator.formula import mean_from_patsy_formula, mean_from_string

def MeanFunction(formula, inputdict={}, use_patsy=True):
    """
    Create a mean function from a formula

    This is the functional interface to creating a mean function from a string formula.
    This method takes a string as an input, an optional dictionary that map strings to
    integer indices in the input data, and an optional boolean flag that indicates if
    the user would like to have the formula parsed with patsy before being converted
    to a mean function.

    The string formulas can be specified in several ways. The formula LHS is implicitly
    always ``"y = "`` or ``"y ~ "``, though these can be explicitly provided as well.
    The RHS may contain a set of terms containing the add, multiply, power, and
    call operations much in the same way that the operations would be entered as
    regular python code. Parentheses are used to indicated prececence as well as
    the call operation, and square brackets indicate an indexing operation on the
    inputs. Inputs may be specified as either a string such as ``"x[0]"``,
    ``"inputs[0]"``, or a string that can be mapped to an integer index with the
    optional dictionary passed to the function. Any strings not representing operations
    or inputs as described above are interpreted as follows: if the string can
    be converted into a number, then it is interpreted as a ``ConstantMean`` fixed
    mean function object; otherwise it is assumed to represent a fitting coefficient.
    Note that this means many characters that do not represent operations within this
    mean function language but would not normally be considered as python variables
    will nonetheless be converted into fitting coefficients -- it is up to the user
    to get this right.

    Expressions that are repeated or redundant will not be simplified, so the user should
    take care that the provided expression is sensible as a mean function and will not
    cause problems when fitting.

    Additional special cases to be aware of:

    * ``call`` cannot be used as a variable name, if this is parsed as a token an exception
      will be raised.
    * ``I`` is the identity operator, it simply returns the given value. It is useful
      if you wish to use patsy to evaluate a formula but protect a part of the string
      formula from being expanded based on the rules in patsy. If ``I`` is encountered
      in any other context, an exception will be raised.

    Examples: ::

        >>> from mogp_emulator.MeanFunction import MeanFunction
        >>> mf1 = MeanFunction("x[0]")
        >>> print(mf1)
        c + c*x[0]
        >>> mf2 = MeanFunction("y = a + b*x[0]", use_patsy=False)
        >>> print(mf2)
        c + c*x[0]
        >>> mf3 = MeanFunction("a*b", {"a": 0, "b": 1})
        >>> print(mf3)
        c + c*x[0] + c*x[1] + c*x[0]*x[1]

    :param formula: string representing the desired mean function formula
    :type formula: str
    :param inputdict: dictionary used to map variables to input indices. Maps
                      strings to integer indices (must be non-negative). Optional,
                      default is ``{}``.
    :type inputdict: dict
    :param use_patsy: Boolean flag indicating if the string is to be parsed using
                      patsy library. Optional, default is ``True``. If patsy is not
                      installed, the basic string parser will be used.
    :type use_patsy: bool
    :returns: New subclass of ``MeanBase`` implementing the given formula
    :rtype: subclass of MeanBase (exact type will depend on the formula that is provided)
    """

    if formula is None or (isinstance(formula, str) and formula.strip() == ""):
        return ConstantMean(0.)

    if not isinstance(formula, str):
        raise ValueError("input formula must be a string")

    if use_patsy:
        mf = mean_from_patsy_formula(formula, inputdict)
    else:
        mf = mean_from_string(formula, inputdict)

    return mf

class MeanBase(object):
    """
    Base mean function class

    The base class for the mean function implementation includes code for checking inputs and
    implements sum, product, power, and composition methods to allow more complicated functions
    to be built up from fixed functions and fitting coefficients. Subclasses need to implement
    the following methods:

    * ``get_n_params`` which returns the number of parameters for a given input size. This is
      usually a constant, but can be more complicated (such as the provided ``PolynomialMean``
      class)
    * ``mean_f`` computes the mean function for the provided inputs and parameters
    * ``mean_deriv`` computes the derivative with respect to the parameters
    * ``mean_hessian`` computes the hessian with respect to the parameters
    * ``mean_inputderiv`` computes the derivate with respect to the inputs

    The base class does not have any attributes, but subclasses will usually have some
    attributes that must be set and so are likely to need a ``__init__`` method.
    """
    def _check_inputs(self, x, params):
        """
        Check the shape of the inputs and reshape if needed

        This method checks that the inputs and parameters are consistent for the provided
        mean function. In particular, the following must be met:

        * The inputs ``x`` must be a 2D numpy array, though if it is 1D it is reshaped to add
          a second dimenion of length 1.
        * ``params`` must be a 1D numpy array. If a multi-dimensional array is provided, it
          will be flattened.
        * ``params`` must have a length that is the same as the return value of ``get_n_params``
          when called with the inputs. Note that some mean functions may have different
          numbers of parameters depending on the inputs, so this may not be known in advance.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: tuple containing the reshaped ``x`` and ``params`` arrays
        :rtype: tuple containing two ndarrays
        """

        x = np.array(x)
        params = np.array(params).flatten()

        if len(x.shape) == 1:
            x = np.reshape(x, (-1, 1))
        assert len(x.shape) == 2, "inputs must be a 1D or 2D array"

        assert len(params.shape) == 1, "params must be a 1D array"

        assert len(params) == self.get_n_params(x), "bad length for params"

        return x, params

    def get_n_params(self, x):
        """
        Determine the number of parameters

        Returns the number of parameters for the mean function, which possibly depends on x.

        :param x: Input array
        :type x: ndarray
        :returns: number of parameters
        :rtype: int
        """

        raise NotImplementedError("base mean function does not implement a particular function")

    def mean_f(self, x, params):
        """
        Returns value of mean function

        Method to compute the value of the mean function for the inputs and parameters provided.
        Shapes of ``x`` and ``params`` must be consistent based on the return value of the
        ``get_n_params`` method. Returns a numpy array of shape ``(x.shape[0],)`` holding
        the value of the mean function for each input point.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function evaluated at all input points, numpy array of shape
                  ``(x.shape[0],)``
        :rtype: ndarray
        """

        raise NotImplementedError("base mean function does not implement a particular function")

    def mean_deriv(self, x, params):
        """
        Returns value of mean function derivative wrt the parameters

        Method to compute the value of the mean function derivative with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, x.shape[0])`` holding the value of the mean
        function derivative with respect to each parameter (first axis) for each input point
        (second axis).

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_params, x.shape[0])``
        :rtype: ndarray
        """

        raise NotImplementedError("base mean function does not implement a particular function")

    def mean_hessian(self, x, params):
        """
        Returns value of mean function Hessian wrt the parameters

        Method to compute the value of the mean function Hessian with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, n_params, x.shape[0])`` holding the value
        of the mean function second derivaties with respect to each parameter pair (first twp axes)
        for each input point (last axis).

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function Hessian with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_parmas, n_params, x.shape[0])``
        :rtype: ndarray
        """

        raise NotImplementedError("base mean function does not implement a particular function")

    def mean_inputderiv(self, x, params):
        """
        Returns value of mean function derivative wrt the inputs

        Method to compute the value of the mean function derivative with respect to the
        inputs for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(x.shape[1], x.shape[0])`` holding the value of the mean
        function derivative with respect to each input (first axis) for each input point
        (second axis).

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the inputs evaluated
                  at all input points, numpy array of shape ``(x.shape[1], x.shape[0])``
        :rtype: ndarray
        """

        raise NotImplementedError("base mean function does not implement a particular function")

    def __add__(self, other):
        """
        Adds two mean functions

        This method adds two mean functions, returning a ``MeanSum`` object. If the second
        argument is a float or integer, it is converted to a ``ConstantMean`` object. If
        the second argument is neither a subclass of ``MeanBase`` nor a float/int,
        an exception is raised.

        :param other: Second ``MeanBase`` (or float/integer) to be added
        :type other: subclass of MeanBase or float or int
        :returns: ``MeanSum`` instance
        :rtype: MeanSum
        """

        if issubclass(type(other), MeanBase):
            return MeanSum(self, other)
        elif isinstance(other, (float, int)):
            return MeanSum(self, ConstantMean(other))
        else:
            raise TypeError("other function cannot be added with a MeanBase")

    def __radd__(self, other):
        """
        Right adds two mean functions

        This method adds two mean functions, returning a ``MeanSum`` object. If the second
        argument is a float or integer, it is converted to a ``ConstantMean`` object. If
        the second argument is neither a subclass of ``MeanBase`` nor a float/int,
        an exception is raised.

        :param other: Second ``MeanBase`` (or float/integer) to be added
        :type other: subclass of MeanBase or float or int
        :returns: ``MeanSum`` instance
        :rtype: MeanSum
        """

        if issubclass(type(other), MeanBase):
            return MeanSum(other, self)
        elif isinstance(other, (float, int)):
            return MeanSum(ConstantMean(other), self)
        else:
            raise TypeError("other function cannot be added with a MeanBase")

    def __mul__(self, other):
        """
        Multiplies two mean functions

        This method multiples two mean functions, returning a ``MeanProduct`` object. If
        the second argument is a float or integer, it is converted to a ``ConstantMean``
        object. If the second argument is neither a subclass of ``MeanBase`` nor a
        float/int, an exception is raised.

        :param other: Second ``MeanBase`` (or float/integer) to be multiplied
        :type other: subclass of MeanBase or float or int
        :returns: ``MeanProduct`` instance
        :rtype: MeanProduct
        """

        if issubclass(type(other), MeanBase):
            return MeanProduct(self, other)
        elif isinstance(other, (float, int)):
            return MeanProduct(self, ConstantMean(other))
        else:
            raise TypeError("other function cannot be multiplied with a MeanBase")

    def __rmul__(self, other):
        """
        Right multiplies two mean functions

        This method multiples two mean functions, returning a ``MeanProduct`` object. If
        the second argument is a float or integer, it is converted to a ``ConstantMean``
        object. If the second argument is neither a subclass of ``MeanBase`` nor a
        float/int, an exception is raised.

        :param other: Second ``MeanBase`` (or float/integer) to be multiplied
        :type other: subclass of MeanBase or float or int
        :returns: ``MeanProduct`` instance
        :rtype: MeanProduct
        """

        if issubclass(type(other), MeanBase):
            return MeanProduct(other, self)
        elif isinstance(other, (float, int)):
            return MeanProduct(ConstantMean(other), self)
        else:
            raise TypeError("other function cannot be multipled with a MeanBase")

    def __pow__(self, exp):
        """
        Raises a mean function to a power

        This method raises a mean function to a power, returning a ``MeanPower`` object.
        The second argument can only be a mean function that returns a value that is
        independent of its input, in particular a ``Coefficient`` or a ``ConstantMean``
        (or a float or integer, from which a new ``ConstantMean`` will be created)
        are the only acceptable types for the ``exp`` argument.

        :param exp: Mean function exponent, must be a ``Coefficient`` or a ``ConstantMean``
                    object, or a float/int from which a new ``ConstantMean`` will be
                    created.
        :type exp: Coefficient, ConstantMean, float, or int
        :returns: ``MeanPower``instance
        :rtype: MeanPower
        """
        if isinstance(exp, (float, int)):
            return MeanPower(self, ConstantMean(exp))
        elif isinstance(exp, (Coefficient, ConstantMean)):
            return MeanPower(self, exp)
        else:
            raise TypeError("MeanBase can only be raised to a power that is a ConstantMean, " +
                            "Coefficient, or float/int")

    def __rpow__(self, base):
        """
        Right raises a mean function to a power

        This method right raises a mean function to a power, meaning that the base
        is potentially not a mean function. Returns a ``MeanPower`` object. The  ``self``
        argument can only be a mean function that returns a value that is
        independent of its input, in particular a ``Coefficient`` or a ``ConstantMean``.
        The base can be any ``MeanBase`` instance, or a float or integer, from which a new
        ``ConstantMean`` will be created.

        :param base: Mean function base, must be a ``MeanBase`` subclass
                     object, or a float/int from which a new ``ConstantMean`` will be
                     created.
        :type base: MeanBase subclass, float, or int
        :returns: ``MeanPower``instance
        :rtype: MeanPower
        """
        if not isinstance(self, (Coefficient, ConstantMean)):
            raise TypeError("arbitrary mean functions cannot serve as the exponent when " +
                            "raising a mean function to a power")
        if isinstance(base, (float, int)):
            return MeanPower(ConstantMean(base), self)
        elif issubclass(type(base), MeanBase):
            return MeanPower(base, self)
        else:
            raise TypeError("base in a MeanPower must be a MeanBase or a float/int")

    def __call__(self, other):
        """
        Composes two mean functions

        This method multiples two mean functions, returning a ``MeanComposite`` object.
        If the second argument is not a subclass of ``MeanBase``, an exception is
        raised.

        :param other: Second ``MeanBase`` to be composed
        :type other: subclass of MeanBase
        :returns: ``MeanComposite`` instance
        :rtype: MeanComposite
        """

        if issubclass(type(other), MeanBase):
            return MeanComposite(self, other)
        else:
            raise TypeError("other function cannot be composed with a MeanBase")

class MeanSum(MeanBase):
    """
    Class representing the sum of two mean functions

    This derived class represents the sum of two mean functions, and does the necessary
    bookkeeping needed to compute the required function and derivatives. The code does
    not do any checks to confirm that it makes sense to add these particular mean functions --
    in particular, adding two ``Coefficient`` classes is the same as having a single
    one, but the code will not attempt to simplify this so it is up to the user to get it
    right.

    :ivar f1: first ``MeanBase`` to be added
    :type f1: subclass of MeanBase
    :ivar f2: second ``MeanBase`` to be added
    :type f2: subclass of MeanBase
    """
    def __init__(self, f1, f2):
        """
        Create a new instance of two added mean functions

        Creates an instance of to added mean functions. Inputs are the two functions
        to be added, which must be subclasses of the base ``MeanBase`` class.

        :param f1: first ``MeanBase`` to be added
        :type f1: subclass of MeanBase
        :param f2: second ``MeanBase`` to be added
        :type f2: subclass of MeanBase
        :returns: new ``MeanSum`` instance
        :rtype: MeanSum
        """
        if not issubclass(type(f1), MeanBase):
            raise TypeError("inputs to MeanSum must be subclasses of MeanBase")
        if not issubclass(type(f2), MeanBase):
            raise TypeError("inputs to MeanSum must be subclasses of MeanBase")

        self.f1 = f1
        self.f2 = f2

    def get_n_params(self, x):
        """
        Determine the number of parameters

        Returns the number of parameters for the mean function, which possibly depends on x.

        :param x: Input array
        :type x: ndarray
        :returns: number of parameters
        :rtype: int
        """
        return self.f1.get_n_params(x) + self.f2.get_n_params(x)

    def mean_f(self, x, params):
        """
        Returns value of mean function

        Method to compute the value of the mean function for the inputs and parameters provided.
        Shapes of ``x`` and ``params`` must be consistent based on the return value of the
        ``get_n_params`` method. Returns a numpy array of shape ``(x.shape[0],)`` holding
        the value of the mean function for each input point.

        For ``MeanSum``, this method applies the sum rule to the results of computing
        the mean for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function evaluated at all input points, numpy array of shape
                  ``(x.shape[0],)``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        return (self.f1.mean_f(x, params[:switch]) +
                self.f2.mean_f(x, params[switch:]))

    def mean_deriv(self, x, params):
        """
        Returns value of mean function derivative wrt the parameters

        Method to compute the value of the mean function derivative with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, x.shape[0])`` holding the value of the mean
        function derivative with respect to each parameter (first axis) for each input point
        (second axis).

        For ``MeanSum``, this method applies the sum rule to the results of computing
        the derivative for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_params, x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        deriv = np.zeros((self.get_n_params(x), x.shape[0]))

        deriv[:switch] = self.f1.mean_deriv(x, params[:switch])
        deriv[switch:] = self.f2.mean_deriv(x, params[switch:])

        return deriv

    def mean_hessian(self, x, params):
        """
        Returns value of mean function Hessian wrt the parameters

        Method to compute the value of the mean function Hessian with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, n_params, x.shape[0])`` holding the value
        of the mean function second derivaties with respect to each parameter pair (first twp axes)
        for each input point (last axis).

        For ``MeanSum``, this method applies the sum rule to the results of computing
        the Hessian for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function Hessian with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_parmas, n_params, x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        hess = np.zeros((self.get_n_params(x), self.get_n_params(x), x.shape[0]))

        hess[:switch, :switch] = self.f1.mean_hessian(x, params[:switch])
        hess[switch:, switch:] = self.f2.mean_hessian(x, params[switch:])

        return hess

    def mean_inputderiv(self, x, params):
        """
        Returns value of mean function derivative wrt the inputs

        Method to compute the value of the mean function derivative with respect to the
        inputs for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(x.shape[1], x.shape[0])`` holding the value of the mean
        function derivative with respect to each input (first axis) for each input point
        (second axis).

        For ``MeanSum``, this method applies the sum rule to the results of computing
        the derivative for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the inputs evaluated
                  at all input points, numpy array of shape ``(x.shape[1], x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        return (self.f1.mean_inputderiv(x, params[:switch]) +
                self.f2.mean_inputderiv(x, params[switch:]))

    def __str__(self):
        """
        Returns a string representation

        Return a formula-like representation of the Mean Function. Useful for confirming
        that a formula was correctly parsed.
        """

        return "{} + {}".format(self.f1, self.f2)

class MeanProduct(MeanBase):
    """
    Class representing the product of two mean functions

    This derived class represents the product of two mean functions, and does the necessary
    bookkeeping needed to compute the required function and derivatives. The code does
    not do any checks to confirm that it makes sense to multiply these particular mean functions --
    in particular, multiplying two ``Coefficient`` classes is the same as having a single
    one, but the code will not attempt to simplify this so it is up to the user to get it
    right.

    :ivar f1: first ``MeanBase`` to be multiplied
    :type f1: subclass of MeanBase
    :ivar f2: second ``MeanBase`` to be multiplied
    :type f2: subclass of MeanBase
    """
    def __init__(self, f1, f2):
        """
        Create a new instance of two mulitplied mean functions

        Creates an instance of to multiplied mean functions. Inputs are the two functions
        to be multiplied, which must be subclasses of the base ``MeanBase`` class.

        :param f1: first ``MeanBase`` to be multiplied
        :type f1: subclass of MeanBase
        :param f2: second ``MeanBase`` to be multiplied
        :type f2: subclass of MeanBase
        :returns: new ``MeanProduct`` instance
        :rtype: MeanProduct
        """

        if not issubclass(type(f1), MeanBase):
            raise TypeError("inputs to MeanProduct must be subclasses of MeanBase")
        if not issubclass(type(f2), MeanBase):
            raise TypeError("inputs to MeanProduct must be subclasses of MeanBase")

        self.f1 = f1
        self.f2 = f2

    def get_n_params(self, x):
        """
        Determine the number of parameters

        Returns the number of parameters for the mean function, which possibly depends on x.

        :param x: Input array
        :type x: ndarray
        :returns: number of parameters
        :rtype: int
        """
        return self.f1.get_n_params(x) + self.f2.get_n_params(x)

    def mean_f(self, x, params):
        """
        Returns value of mean function

        Method to compute the value of the mean function for the inputs and parameters provided.
        Shapes of ``x`` and ``params`` must be consistent based on the return value of the
        ``get_n_params`` method. Returns a numpy array of shape ``(x.shape[0],)`` holding
        the value of the mean function for each input point.

        For ``MeanProduct``, this method applies the product rule to the results of computing
        the mean for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function evaluated at all input points, numpy array of shape
                  ``(x.shape[0],)``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        return (self.f1.mean_f(x, params[:switch])*
                self.f2.mean_f(x, params[switch:]))

    def mean_deriv(self, x, params):
        """
        Returns value of mean function derivative wrt the parameters

        Method to compute the value of the mean function derivative with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, x.shape[0])`` holding the value of the mean
        function derivative with respect to each parameter (first axis) for each input point
        (second axis).

        For ``MeanProduct``, this method applies the product rule to the results of computing
        the derivative for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_params, x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        deriv = np.zeros((self.get_n_params(x), x.shape[0]))

        deriv[:switch] = (self.f1.mean_deriv(x, params[:switch])*
                          self.f2.mean_f(x, params[switch:]))

        deriv[switch:] = (self.f1.mean_f(x, params[:switch])*
                          self.f2.mean_deriv(x, params[switch:]))

        return deriv

    def mean_hessian(self, x, params):
        """
        Returns value of mean function Hessian wrt the parameters

        Method to compute the value of the mean function Hessian with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, n_params, x.shape[0])`` holding the value
        of the mean function second derivaties with respect to each parameter pair (first twp axes)
        for each input point (last axis).

        For ``MeanProduct``, this method applies the product rule to the results of computing
        the Hessian for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function Hessian with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_parmas, n_params, x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        hess = np.zeros((self.get_n_params(x), self.get_n_params(x), x.shape[0]))

        hess[:switch, :switch] = (self.f1.mean_hessian(x, params[:switch])*
                                  self.f2.mean_f(x, params[switch:]))
        hess[:switch, switch:, :] = (self.f1.mean_deriv(x, params[:switch])[:,np.newaxis,:]*
                                     self.f2.mean_deriv(x, params[switch:])[np.newaxis,:,:])
        hess[switch:, :switch, :] = np.transpose(hess[:switch, switch:, :], (1, 0, 2))
        hess[switch:, switch:] = (self.f1.mean_f(x, params[:switch])*
                                  self.f2.mean_hessian(x, params[switch:]))

        return hess


    def mean_inputderiv(self, x, params):
        """
        Returns value of mean function derivative wrt the inputs

        Method to compute the value of the mean function derivative with respect to the
        inputs for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(x.shape[1], x.shape[0])`` holding the value of the mean
        function derivative with respect to each input (first axis) for each input point
        (second axis).

        For ``MeanProduct``, this method applies the product rule to the results of computing
        the derivative for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the inputs evaluated
                  at all input points, numpy array of shape ``(x.shape[1], x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        return (self.f1.mean_inputderiv(x, params[:switch])*
                self.f2.mean_f(x, params[switch:]) +
                self.f1.mean_f(x, params[:switch])*
                self.f2.mean_inputderiv(x, params[switch:]))

    def __str__(self):
        """
        Returns a string representation

        Return a formula-like representation of the Mean Function. Useful for confirming
        that a formula was correctly parsed.
        """

        return "{}*{}".format(self.f1, self.f2)

class MeanPower(MeanBase):
    """
    Class representing a mean function raised to a power

    This derived class represents a mean function raised to a power, and does the necessary
    bookkeeping needed to compute the required function and derivatives. The code requires
    that the exponent be either a ``Coefficient``, ``ConstantMean``, ``float``, or ``int``
    as the output of the exponent mean function must be independent of the inputs to make
    sense. If input is a float or int, a ``ConstantMean`` instance will be created.

    :ivar f1: first ``MeanBase`` to be raised to the given exponent
    :type f1: subclass of MeanBase
    :ivar f2: second ``MeanBase`` indicating the exponent. Must be a ``Coefficient``,
              ``ConstantMean``, or float/int (from which a ``ConstantMean`` object will
              be created)
    :type f2: Coefficient, ConstantMean, float, or int
    """
    def __init__(self, f1, f2):
        """
        Create a new instance of a mean function raised to a power

        Creates an instance of a mean function raised to a power. Inputs are the two
        functions (base, exponent), the first of which must be subclass of the base
        ``MeanBase`` class, and the second must be a ``Coefficient`` or a
        ``ConstantMean`` (or a float or int, from which a ``ConstantMean`` will
        be created).

        :param f1: first ``MeanBase`` serving as the base
        :type f1: subclass of MeanBase
        :param f2: second ``MeanBase`` serving as the exponent, must be a ``Coefficient``,
                   ``ConstantMean``, ``float``, or ``int``
        :type f2: Coefficient, ConstantMean, float, or int
        :returns: new ``MeanPower`` instance
        :rtype: MeanPower
        """

        if not issubclass(type(f1), MeanBase):
            raise TypeError("first input to MeanPower must be a subclass of MeanBase")
        if isinstance(f2, (float, int)):
            f2 = ConstantMean(f2)
        if not isinstance(f2, (ConstantMean, Coefficient)):
            raise TypeError("second input to MeanPower must be a Coefficient, ConstantMean, "
                            "float, or int")

        self.f1 = f1
        self.f2 = f2

    def get_n_params(self, x):
        """
        Determine the number of parameters

        Returns the number of parameters for the mean function, which possibly depends on x.

        :param x: Input array
        :type x: ndarray
        :returns: number of parameters
        :rtype: int
        """
        return self.f1.get_n_params(x) + self.f2.get_n_params(x)

    def mean_f(self, x, params):
        """
        Returns value of mean function

        Method to compute the value of the mean function for the inputs and parameters provided.
        Shapes of ``x`` and ``params`` must be consistent based on the return value of the
        ``get_n_params`` method. Returns a numpy array of shape ``(x.shape[0],)`` holding
        the value of the mean function for each input point.

        For ``MeanProduct``, this method applies the product rule to the results of computing
        the mean for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function evaluated at all input points, numpy array of shape
                  ``(x.shape[0],)``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        return (self.f1.mean_f(x, params[:switch])**
                self.f2.mean_f(x, params[switch:]))

    def mean_deriv(self, x, params):
        """
        Returns value of mean function derivative wrt the parameters

        Method to compute the value of the mean function derivative with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, x.shape[0])`` holding the value of the mean
        function derivative with respect to each parameter (first axis) for each input point
        (second axis).

        For ``MeanPpwer``, this method applies the power rule to the results of computing
        the derivative for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_params, x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        exp = self.f2.mean_f(x, params[switch:])
        nonzeroexp = True
        if np.allclose(exp, 0.):
            nonzeroexp = False

        deriv = np.zeros((self.get_n_params(x), x.shape[0]))

        if nonzeroexp:
            deriv[:switch] = (exp*self.f1.mean_f(x, params[:switch])**(exp - 1.)*
                              self.f1.mean_deriv(x, params[:switch]))

        # only evaluate if f2 has parameters, as f1 could be negative and taking the log will
        # raise an error even though this calculation is ultimately ignored in this case

        if not self.f2.get_n_params(x) == 0:

            deriv[switch:] = (np.log(self.f1.mean_f(x, params[:switch]))*
                              self.f1.mean_f(x, params[:switch])**exp*
                              self.f2.mean_deriv(x, params[switch:]))

        return deriv

    def mean_hessian(self, x, params):
        """
        Returns value of mean function Hessian wrt the parameters

        Method to compute the value of the mean function Hessian with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, n_params, x.shape[0])`` holding the value
        of the mean function second derivaties with respect to each parameter pair (first twp axes)
        for each input point (last axis).

        For ``MeanPower``, this method applies the power rule to the results of computing
        the Hessian for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function Hessian with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_parmas, n_params, x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        exp = self.f2.mean_f(x, params[switch:])
        nonzeroexp = True
        if np.allclose(exp, 0.):
            nonzeroexp = False
        nononeexp = True
        if np.allclose(exp, 1.):
            nononeexp = False

        hess = np.zeros((self.get_n_params(x), self.get_n_params(x), x.shape[0]))

        if nonzeroexp and nononeexp:

            hess[:switch, :switch] = (exp*self.f1.mean_f(x, params[:switch])**(exp - 1.)*
                                      self.f1.mean_hessian(x, params[:switch]) +
                                      exp*(exp - 1.)*self.f1.mean_f(x, params[:switch])**(exp - 2.)*
                                      self.f1.mean_deriv(x, params[:switch]))

        elif nonzeroexp:

            hess[:switch, :switch] = (exp*self.f1.mean_f(x, params[:switch])**(exp - 1.)*
                                      self.f1.mean_hessian(x, params[:switch]))


        if not self.f2.get_n_params(x) == 0:

            if nonzeroexp:
                hess[:switch, switch:, :] = (self.f1.mean_f(x, params[:switch])**(exp - 1.)*
                                             (exp*np.log(self.f1.mean_f(x, params[:switch])) + 1.)*
                                             self.f1.mean_deriv(x, params[:switch])[:,np.newaxis,:]*
                                             self.f2.mean_deriv(x, params[switch:])[np.newaxis,:,:])
                hess[switch:, :switch, :] = np.transpose(hess[:switch, switch:, :], (1, 0, 2))

            hess[switch:, switch:] = (self.f1.mean_f(x, params[:switch])**exp*
                                      (np.log(self.f1.mean_f(x, params[:switch]))**2*
                                       self.f2.mean_deriv(x, params[switch:])**2 +
                                       np.log(self.f1.mean_f(x, params[:switch]))*
                                       self.f2.mean_hessian(x, params[switch:])))

        return hess


    def mean_inputderiv(self, x, params):
        """
        Returns value of mean function derivative wrt the inputs

        Method to compute the value of the mean function derivative with respect to the
        inputs for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(x.shape[1], x.shape[0])`` holding the value of the mean
        function derivative with respect to each input (first axis) for each input point
        (second axis).

        For ``MeanPower``, this method applies the power rule to the results of computing
        the derivative for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the inputs evaluated
                  at all input points, numpy array of shape ``(x.shape[1], x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        exp = self.f2.mean_f(x, params[switch:])
        nonzeroexp = True
        if np.allclose(exp, 0.):
            nonzeroexp = False

        inputderiv = np.zeros((x.shape[1], x.shape[0]))

        if nonzeroexp:
            inputderiv = (exp*self.f1.mean_f(x, params[:switch])**(exp - 1.)*
                          self.f1.mean_inputderiv(x, params[:switch]))

        return inputderiv

    def __str__(self):
        """
        Returns a string representation

        Return a formula-like representation of the Mean Function. Useful for confirming
        that a formula was correctly parsed.
        """

        return "{}^{}".format(self.f1, self.f2)

class MeanComposite(MeanBase):
    """
    Class representing the composition of two mean functions

    This derived class represents the composition of two mean functions, and does the necessary
    bookkeeping needed to compute the required function and derivatives. The code does
    not do any checks to confirm that it makes sense to compose these particular mean
    functions -- in particular, applying a ``Coefficient`` class to another function will
    simply wipe out the second function. This will not raise an error, but the code will not
    attempt to alert the user to this so it is up to the user to get it right.

    Because the Hessian computation requires mixed partials that are not normally implemented
    in the ``MeanBase`` class, the Hessian computation is not currently implemented.
    If you require Hessian computation for a composite mean function, you must implement
    it yourself.

    Note that since the outer function takes as its input the output of the second function,
    the outer function can only ever have an index of 0 due to the fixed output shape of
    a mean function. This will produce an error when attempting to evaluate the function
    or its derivatives, but will not cause an error when initializing a ``MeanComposite``
    object.

    :ivar f1: first ``MeanBase`` to be applied to the second
    :type f1: subclass of MeanBase
    :ivar f2: second ``MeanBase`` to be composed as the input to the first
    :type f2: subclass of MeanBase
    """
    def __init__(self, f1, f2):
        """
        Create a new instance of two composed mean functions

        Creates an instance of to composed mean functions. Inputs are the two functions
        to be composed (``f1(f2)``), which must be subclasses of the base ``MeanBase``
        class.

        :param f1: first ``MeanBase`` to be applied to the second
        :type f1: subclass of MeanBase
        :param f2: second ``MeanBase`` to be composed as the input to the first
        :type f2: subclass of MeanBase
        :returns: new ``MeanComposite`` instance
        :rtype: MeanComposite
        """
        if not issubclass(type(f1), MeanBase):
            raise TypeError("inputs to MeanComposite must be subclasses of MeanBase")
        if not issubclass(type(f2), MeanBase):
            raise TypeError("inputs to MeanComposite must be subclasses of MeanBase")

        self.f1 = f1
        self.f2 = f2

    def get_n_params(self, x):
        """
        Determine the number of parameters

        Returns the number of parameters for the mean function, which possibly depends on x.

        :param x: Input array
        :type x: ndarray
        :returns: number of parameters
        :rtype: int
        """
        return self.f1.get_n_params(np.zeros((x.shape[0], 1))) + self.f2.get_n_params(x)

    def mean_f(self, x, params):
        """
        Returns value of mean function

        Method to compute the value of the mean function for the inputs and parameters provided.
        Shapes of ``x`` and ``params`` must be consistent based on the return value of the
        ``get_n_params`` method. Returns a numpy array of shape ``(x.shape[0],)`` holding
        the value of the mean function for each input point.

        For ``MeanComposite``, this method applies the output of the second function as
        input to the first function.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function evaluated at all input points, numpy array of shape
                  ``(x.shape[0],)``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        return self.f1.mean_f(np.reshape(self.f2.mean_f(x, params[switch:]), (-1, 1)),
                              params[:switch])

    def mean_deriv(self, x, params):
        """
        Returns value of mean function derivative wrt the parameters

        Method to compute the value of the mean function derivative with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, x.shape[0])`` holding the value of the mean
        function derivative with respect to each parameter (first axis) for each input point
        (second axis).

        For ``MeanComposite``, this method applies the chain rule to the results of computing
        the derivative for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_params, x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        deriv = np.zeros((self.get_n_params(x), x.shape[0]))

        f2 = np.reshape(self.f2.mean_f(x, params[switch:]), (-1, 1))

        deriv[:switch] = self.f1.mean_deriv(f2, params[:switch])

        deriv[switch:] = (self.f1.mean_inputderiv(f2, params[:switch])*
                          self.f2.mean_deriv(x, params[switch:]))

        return deriv

    def mean_inputderiv(self, x, params):
        """
        Returns value of mean function derivative wrt the inputs

        Method to compute the value of the mean function derivative with respect to the
        inputs for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(x.shape[1], x.shape[0])`` holding the value of the mean
        function derivative with respect to each input (first axis) for each input point
        (second axis).

        For ``MeanComposite``, this method applies the chain rule to the results of computing
        the derivative for the individual functions.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the inputs evaluated
                  at all input points, numpy array of shape ``(x.shape[1], x.shape[0])``
        :rtype: ndarray
        """

        switch = self.f1.get_n_params(x)

        return (self.f1.mean_inputderiv(np.reshape(self.f2.mean_f(x, params[switch:]), (-1, 1)),
                                        params[:switch])*
                self.f2.mean_inputderiv(x, params[switch:]))

    def __str__(self):
        """
        Returns a string representation

        Return a formula-like representation of the Mean Function. Useful for confirming
        that a formula was correctly parsed.
        """

        return "{}({})".format(self.f1, self.f2)

class FixedMean(MeanBase):
    """
    Class representing a fixed mean function with no parameters

    Class representing a mean function with a fixed function (and optional derivative)
    and no fitting parameters. The user must provide these functions when initializing
    the instance.

    :ivar f: fixed mean function, must be callable and take a single argument (the inputs)
    :type f: function
    :ivar deriv: fixed derivative function (optional if no derivatives are needed), must
                   be callable and take a single argument (the inputs)
    :type deriv: function or None
    """
    def __init__(self, f, deriv=None):
        """
        Initialize a class instance representing a fixed mean function with no parameters

        Create a class instance representing a mean function with a fixed function
        (and optional derivative) and no fitting parameters. The user must provide these
        functions, though the derivative is optional. The code will check that the provided
        arguments are callable, but will not confirm that the inputs and outputs are the
        correct type/shape.

        :param f: fixed mean function, must be callable and take a single argument (the inputs)
        :type f: function
        :param deriv: fixed derivative function (optional if no derivatives are needed), must
                       be callable and take a single argument (the inputs)
        :type deriv: function or None
        :returns: new ``FixedMean`` instance
        :rtype: FixedMean
        """
        assert callable(f), "fixed mean function must be a callable function"
        if not deriv is None:
            assert callable(deriv), "mean function derivative must be a callable function"

        self.f = f
        self.deriv = deriv

    def get_n_params(self, x):
        """
        Determine the number of parameters

        Returns the number of parameters for the mean function, which possibly depends on x.
        For a ``FixedMean`` class, this is zero.

        :param x: Input array
        :type x: ndarray
        :returns: number of parameters
        :rtype: int
        """
        return 0

    def mean_f(self, x, params):
        """
        Returns value of mean function

        Method to compute the value of the mean function for the inputs and parameters provided.
        Shapes of ``x`` and ``params`` must be consistent based on the return value of the
        ``get_n_params`` method. For ``FixedMean`` classes, there are no parameters so the
        ``params`` argument should be an array of length zero. Returns a numpy array of shape
        ``(x.shape[0],)`` holding the value of the mean function for each input point.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input (zero in this case)
        :type params: ndarray
        :returns: Value of mean function evaluated at all input points, numpy array of shape
                  ``(x.shape[0],)``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        return self.f(x)

    def mean_deriv(self, x, params):
        """
        Returns value of mean function derivative wrt the parameters

        Method to compute the value of the mean function derivative with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        For ``FixedMean`` classes, there are no parameters so the ``params`` argument
        should be an array of length zero. Returns a numpy array of shape
        ``(n_params, x.shape[0])`` holding the value of the mean function derivative with
        respect to each parameter (first axis) for each input point (second axis). Since
        fixed means have no parameters, this will just be an array of zeros.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_params, x.shape[0])``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        return np.zeros((self.get_n_params(x), x.shape[0]))

    def mean_hessian(self, x, params):
        """
        Returns value of mean function Hessian wrt the parameters

        Method to compute the value of the mean function Hessian with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        For ``FixedMean`` classes, there are no parameters so the ``params`` argument
        should be an array of length zero. Returns a numpy array of shape
        ``(n_params, n_params, x.shape[0])`` holding the value of the mean function
        second derivaties with respect to each parameter pair (first twp axes) for each
        input point (last axis). Since fixed means have no parameters, this will just
        be an array of zeros.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function Hessian with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_parmas, n_params, x.shape[0])``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        return np.zeros((self.get_n_params(x), self.get_n_params(x), x.shape[0]))

    def mean_inputderiv(self, x, params):
        """
        Returns value of mean function derivative wrt the inputs

        Method to compute the value of the mean function derivative with respect to the
        inputs for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        For ``FixedMean`` classes, there are no parameters so the ``params`` argument
        should be an array of length zero. Returns a numpy array of shape
        ``(x.shape[1], x.shape[0])`` holding the value of the mean function derivative
        with respect to each input (first axis) for each input point (second axis).

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the inputs evaluated
                  at all input points, numpy array of shape ``(x.shape[1], x.shape[0])``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        if self.deriv is None:
            raise RuntimeError("Derivative function was not provided with this FixedMean")
        else:
            return self.deriv(x)

    def __str__(self):
        """
        Returns a string representation

        Return a formula-like representation of the Mean Function. Useful for confirming
        that a formula was correctly parsed.
        """

        return "f"

def fixed_f(x, index, f):
    """
    Dummy function to index into x and apply a function

    Usage is intended to be with a fixed mean function, where an index and specific mean
    function are meant to be bound using partial before setting it as the ``f`` attribute of
    ``FixedMean``

    :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
    :type x: ndarray
    :param index: integer index to be applied to the second axis of ``x``, used to select
                  a particular input variable. Must be non-negative and less than the
                  length of the second axis of the inputs.
    :type index: int
    :param f: fixed mean function, must be callable and take a single argument (the inputs)
    :type f: function
    :returns: Value of mean function evaluated at all input points, numpy array of shape
              ``(x.shape[0],)``
    :rtype: ndarray
    """
    assert callable(f), "fixed mean function must be callable"
    assert index >= 0, "provided index cannot be negative"
    assert x.ndim == 2, "x must have 2 dimensions"

    try:
        val = f(x[:,index])
    except IndexError:
        raise IndexError("provided mean function index is out of range")

    return val

def fixed_inputderiv(x, index, deriv):
    """
    Dummy function to index into x and apply a derivative function

    Usage is intended to be with a fixed mean function, where an index and specific derivative
    function are meant to be bound using partial before setting it as the ``deriv`` attribute of
    ``FixedMean``

    :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
    :type x: ndarray
    :param index: integer index to be applied to the second axis of ``x``, used to select
                  a particular input variable. Must be non-negative and less than the
                  length of the second axis of the inputs.
    :type index: int
    :param deriv: fixed derivative function, must be callable and take a single argument (the inputs)
    :type deriv: function
    :returns: Value of mean derivative evaluated at all input points, numpy array of shape
              ``(x.shape[1], x.shape[0])``
    :rtype: ndarray
    """
    assert callable(deriv), "fixed mean function derivative must be callable"
    assert index >= 0, "provided index cannot be negative"
    assert x.ndim == 2, "x must have 2 dimensions"

    try:
        out = np.zeros((x.shape[1], x.shape[0]))
        out[index, :] = deriv(np.transpose(x[:, index]))
    except IndexError:
        raise IndexError("provided mean function index is out of range")

    return out

def one(x):
    """
    Function to return an array of ones with the same shape as the input

    Function to return a numpy array of ones with the same shape as the input. Used in
    linear mean functions to evaluate derivatives.

    :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
    :type x: ndarray
    :returns: Numpy array of ones with the same shape as x
    :rtype: ndarray
    """
    return np.ones(x.shape)

def const_f(x, val):
    """
    Function to return an array of a constant value

    Function to return a numpy array of a constant value with the correct shape for a given
    input. Used in constant mean functions to evaluate the function.

    :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
    :type x: ndarray
    :param val: value of output, must be a float
    :type val: float
    :returns: Numpy array of ``val`` with shape ``(x.shape[0],)``
    :rtype: ndarray
    """
    assert x.ndim == 2, "x must have 2 dimensions"

    return np.broadcast_to(val, x.shape[0])

def const_deriv(x):
    """
    Function to return an array of zeros with the transposed shape of the inputs

    Function to return a numpy array of zeros with the shape that is transpose of the
    shape of the input. Used in constant mean functions to evaluate the derivative.

    :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
    :type x: ndarray
    :returns: Numpy array of zeros with shape ``(x.shape[1], x.shape[0])``
    :rtype: ndarray
    """
    assert x.ndim == 2, "x must have 2 dimensions"

    return np.zeros((x.shape[1], x.shape[0]))

class ConstantMean(FixedMean):
    """
    Class representing a constant fixed mean function

    Subclass of ``FixedMean`` where the function is a constant, with the value
    provided when ``ConstantMean`` is initialized. Uses utility functions to bind the
    value to the ``fixed_f`` function and sets that as the ``f`` attribute.

    :ivar f: fixed mean function, must be callable and take a single argument (the inputs)
    :type f: function
    :ivar deriv: fixed derivative function (optional if no derivatives are needed), must
                   be callable and take a single argument (the inputs)
    :type deriv: function
    """
    def __init__(self, val):
        """
        Initialize a new ConstantMean

        Create a new ``ConstantMean`` instance with the given constant value.

        :param val: Constant mean function value, must be a float or an integer
        :type val: float or int
        :returns: new ``ConstantMean`` instance
        :rtype: ConstantMean
        """
        if not isinstance(val, (float, int)):
            raise TypeError("val must be a float or an integer")
        self.f = partial(const_f, val=val)
        self.deriv = const_deriv

    def __str__(self):
        """
        Returns a string representation

        Return a formula-like representation of the Mean Function. Useful for confirming
        that a formula was correctly parsed.
        """
        val = signature(self.f).parameters['val'].default
        return "{}".format(val)

class LinearMean(FixedMean):
    """
    Class representing a linear fixed mean function

    Subclass of ``FixedMean`` where the function is a linear function. By default the
    function is linear in the first input dimension, though any non-negative integer index
    can be provided to control which input is used in the linear function. Uses utility
    functions to bind the correct function to the ``fixed_f`` function and sets that as
    the ``f`` attribute and similar with the ``fixed_deriv`` utility function and the
    ``deriv`` attribute.

    :ivar f: fixed mean function, must be callable and take a single argument (the inputs)
    :type f: function
    :ivar deriv: fixed derivative function, must be callable and take a single argument
                   (the inputs)
    :type deriv: function
    """
    def __init__(self, index=0):
        """
        Initialize a new LinearMean

        Create a new ``LinearMean`` instance with the given index value. This index is used
        to select the dimension of the input for evaluating the function.

        :param index: integer index to be applied to the second axis of ``x``, used to select
                      a particular input variable. Must be non-negative and less than the
                      length of the second axis of the inputs.
        :type index: int
        :returns: new ``LinearMean`` instance
        :rtype: LinearMean
        """
        self.f = partial(fixed_f, index=index, f=np.array)
        self.deriv = partial(fixed_inputderiv, index=index, deriv=one)

    def __str__(self):
        """
        Returns a string representation

        Return a formula-like representation of the Mean Function. Useful for confirming
        that a formula was correctly parsed.
        """

        index = signature(self.f).parameters["index"].default

        return "x[{}]".format(index)

class Coefficient(MeanBase):
    """
    Class representing a single fitting parameter in a mean function

    Class representing a mean function with single free fitting parameter. Does not require any
    internal state as the parameter value is stored/set externally through fitting routines.
    """
    def get_n_params(self, x):
        """
        Determine the number of parameters

        Returns the number of parameters for the mean function, which possibly depends on x.
        For a ``Coefficient`` class, this is always 1.

        :param x: Input array
        :type x: ndarray
        :returns: number of parameters
        :rtype: int
        """
        return 1

    def mean_f(self, x, params):
        """
        Returns value of mean function

        Method to compute the value of the mean function for the inputs and parameters provided.
        Shapes of ``x`` and ``params`` must be consistent based on the return value of the
        ``get_n_params`` method. For ``Coefficient`` classes, the inputs are ignored and the
        function returns the value of the parameter broadcasting it appropriately given the
        shape of the inputs. Thus, the ``params`` argument should always be an array of length
        one. Returns a numpy array of shape ``(x.shape[0],)`` holding the value of the
        parameter for each input point.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input (one in this case)
        :type params: ndarray
        :returns: Value of mean function evaluated at all input points, numpy array of shape
                  ``(x.shape[0],)``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        return np.broadcast_to(params, x.shape[0])

    def mean_deriv(self, x, params):
        """
        Returns value of mean function derivative wrt the parameters

        Method to compute the value of the mean function derivative with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        For ``Coefficient`` classes, the inputs are ignored and the derivative function
        returns one, broadcasting it appropriately given the shape of the inputs.
        Returns a numpy array of ones with shape ``(1, x.shape[0])`` holding the value
        of the mean function derivative with respect to each parameter (first axis) for
        each input point (second axis). Since coefficients are single parameters, this
        will just be an array of ones.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_params, x.shape[0])``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        return np.ones((self.get_n_params(x), x.shape[0]))

    def mean_hessian(self, x, params):
        """
        Returns value of mean function Hessian wrt the parameters

        Method to compute the value of the mean function Hessian with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        For ``Coefficient`` classes, there is only a single parameter so the ``params``
        argument should be an array of length one. Returns a numpy array of shape
        ``(n_params, n_params, x.shape[0])`` holding the value of the mean function
        second derivaties with respect to each parameter pair (first twp axes) for each
        input point (last axis). Since coefficients depend linearly on a single parameter,
        this will always be an array of zeros.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function Hessian with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_parmas, n_params, x.shape[0])``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        return np.zeros((self.get_n_params(x), self.get_n_params(x), x.shape[0]))

    def mean_inputderiv(self, x, params):
        """
        Returns value of mean function derivative wrt the inputs

        Method to compute the value of the mean function derivative with respect to the
        inputs for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        For ``Coefficient`` classes, there is a single parameters so the ``params`` argument
        should be an array of length one. Returns a numpy array of shape
        ``(x.shape[1], x.shape[0])`` holding the value of the mean function derivative
        with respect to each input (first axis) for each input point (second axis).
        Since coefficients do not depend on the inputs, this is just an array of zeros.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the inputs evaluated
                  at all input points, numpy array of shape ``(x.shape[1], x.shape[0])``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        return np.zeros((x.shape[1], x.shape[0]))

    def __str__(self):
        """
        Returns a string representation

        Return a formula-like representation of the Mean Function. Useful for confirming
        that a formula was correctly parsed.
        """
        return "c"

class PolynomialMean(MeanBase):
    """
    Polynomial mean function class

    A ``PolynomialMean`` is a mean function where every input dimension is fit to a fixed
    degree polynomial. The degree must be provided when creating the class instance. The
    number of parameters depends on the degree and the shape of the inputs, since a separate
    set of parameters are used for each input dimension.

    :ivar degree: Polynomial degree, must be a positive integer
    :type degree: int
    """
    def __init__(self, degree):
        """
        Create a new polynomial mean function instance

        A ``PolynomialMean`` is a mean function where every input dimension is fit to a fixed
        degree polynomial. The degree must be provided when creating the class instance. The
        number of parameters depends on the degree and the shape of the inputs, since a separate
        set of parameters are used for each input dimension. Must provide the degree when
        initializing.

        :param degree: Polynomial degree, must be a positive integer
        :type degree: int
        :returns: new ``PolynomialMean`` instance
        :rtype: PolynomialMean
        """
        assert int(degree) > 0, "degree must be a positive integer"

        self.degree = int(degree)

    def get_n_params(self, x):
        """
        Determine the number of parameters

        Returns the number of parameters for the mean function, which depends on x.

        :param x: Input array
        :type x: ndarray
        :returns: number of parameters
        :rtype: int
        """
        return x.shape[1]*self.degree + 1

    def mean_f(self, x, params):
        """
        Returns value of mean function

        Method to compute the value of the mean function for the inputs and parameters provided.
        Shapes of ``x`` and ``params`` must be consistent based on the return value of the
        ``get_n_params`` method. Returns a numpy array of shape ``(x.shape[0],)`` holding
        the value of the mean function for each input point.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function evaluated at all input points, numpy array of shape
                  ``(x.shape[0],)``
        :rtype: ndarray
        """
        x, params = self._check_inputs(x, params)

        n_params = self.get_n_params(x)

        indices = np.arange(0, n_params - 1) % x.shape[1]
        expon = np.arange(0, n_params - 1) // x.shape[1] + 1

        output = params[0] + np.sum(params[1:]*x[:, indices]**expon, axis = 1)

        return output

    def mean_deriv(self, x, params):
        """
        Returns value of mean function derivative wrt the parameters

        Method to compute the value of the mean function derivative with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, x.shape[0])`` holding the value of the mean
        function derivative with respect to each parameter (first axis) for each input point
        (second axis).

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_params, x.shape[0])``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        n_params = self.get_n_params(x)

        deriv = np.zeros((n_params, x.shape[0]))
        deriv[0] = np.ones(x.shape[0])

        indices = np.arange(0, n_params - 1) % x.shape[1]
        expon = np.arange(0, n_params - 1) // x.shape[1] + 1

        deriv[1:,:] = np.transpose(x[:, indices]**expon)

        return deriv

    def mean_hessian(self, x, params):
        """
        Returns value of mean function Hessian wrt the parameters

        Method to compute the value of the mean function Hessian with respect to the
        parameters for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(n_params, n_params, x.shape[0])`` holding the
        value of the mean function second derivaties with respect to each parameter pair
        (first twp axes) for each input point (last axis). Since polynomial means depend
        linearly on all input parameters, this will always be an array of zeros.

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function Hessian with respect to the parameters evaluated
                  at all input points, numpy array of shape ``(n_parmas, n_params, x.shape[0])``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        n_params = self.get_n_params(x)

        hess = np.zeros((n_params, n_params, x.shape[0]))

        return hess

    def mean_inputderiv(self, x, params):
        """
        Returns value of mean function derivative wrt the inputs

        Method to compute the value of the mean function derivative with respect to the
        inputs for the inputs and parameters provided. Shapes of ``x`` and ``params``
        must be consistent based on the return value of the ``get_n_params`` method.
        Returns a numpy array of shape ``(x.shape[1], x.shape[0])`` holding the value of the mean
        function derivative with respect to each input (first axis) for each input point
        (second axis).

        :param x: Inputs, must be a 1D or 2D numpy array (if 1D a second dimension will be added)
        :type x: ndarray
        :param params: Parameters, must be a 1D numpy array (of more than 1D will be flattened)
                       and have the same length as the number of parameters required for the
                       provided input
        :type params: ndarray
        :returns: Value of mean function derivative with respect to the inputs evaluated
                  at all input points, numpy array of shape ``(x.shape[1], x.shape[0])``
        :rtype: ndarray
        """

        x, params = self._check_inputs(x, params)

        expon = np.reshape(np.arange(0, x.shape[0]*x.shape[1]*self.degree)//x.shape[1]//self.degree,
                           (self.degree, x.shape[0]*x.shape[1]))
        x_indices = np.reshape(np.arange(0, x.shape[0]*x.shape[1]*self.degree) % (x.shape[0]*x.shape[1]),
                               (self.degree, x.shape[0]*x.shape[1]))
        param_indices = np.reshape(np.arange(0, x.shape[0]*x.shape[1]*self.degree) % x.shape[1],
                                   (self.degree, x.shape[0]*x.shape[1])) + expon*x.shape[1]
        param_indices = np.reshape(param_indices, (self.degree, x.shape[0]*x.shape[1]))

        output = np.sum((expon + 1.)*params[1:][param_indices]*x.flatten()[x_indices]**expon, axis=0)

        return np.transpose(np.reshape(output, (x.shape[0], x.shape[1])))

    def __str__(self):
        """
        Returns a string representation

        Return a string representation of the polynomial mean
        """

        return "Polynomial mean of degree {}".format(self.degree)



from tempfile import TemporaryFile
import numpy as np
import pytest
from numpy.testing import assert_allclose
from ..GaussianProcess import GaussianProcess, PredictResult
from ..MeanFunction import ConstantMean, LinearMean, MeanFunction
from ..Kernel import SquaredExponential, Matern52
from ..Priors import NormalPrior, GammaPrior, InvGammaPrior
from scipy import linalg

@pytest.fixture
def x():
    return np.array([[1., 2., 3.], [4., 5., 6.]])

@pytest.fixture
def y():
    return np.array([2., 4.])

def test_GaussianProcess_init(x, y):
    "Test function for correct functioning of the init method of GaussianProcess"

    gp = GaussianProcess(x, y)
    assert_allclose(x, gp.inputs)
    assert_allclose(y, gp.targets)
    assert gp.D == 3
    assert gp.n == 2
    assert gp.nugget == None

    gp = GaussianProcess(y, y)
    assert gp.inputs.shape == (2, 1)

    gp = GaussianProcess(x, y, nugget=1.e-12)
    assert_allclose(gp.nugget, 1.e-12)

    gp = GaussianProcess(x, y, mean=ConstantMean(1.), kernel=Matern52(), nugget="fit")

    gp = GaussianProcess(x, y, kernel="SquaredExponential")
    assert isinstance(gp.kernel, SquaredExponential)

    gp = GaussianProcess(x, y, kernel="Matern52")
    assert isinstance(gp.kernel, Matern52)

    gp = GaussianProcess(x, y, mean="a+b*x[0]", use_patsy=False)

    assert str(gp.mean) == "c + c*x[0]"

    gp = GaussianProcess(x, y, mean="c", inputdict={"c": 0})

    assert str(gp.mean) == "c + c*x[0]"

def test_GP_init_failures(x, y):
    "Tests that GaussianProcess fails correctly with bad inputs"

    with pytest.raises(AssertionError):
        gp = GaussianProcess(np.ones((2, 2, 2)), y)

    with pytest.raises(AssertionError):
        gp = GaussianProcess(x, x)

    with pytest.raises(AssertionError):
        gp = GaussianProcess(np.ones((2, 3)), np.ones(3))

    with pytest.raises(ValueError):
        gp = GaussianProcess(x, y, mean=1)

    with pytest.raises(ValueError):
        gp = GaussianProcess(x, y, kernel="blah")

    with pytest.raises(ValueError):
        gp = GaussianProcess(x, y, kernel=1)

    with pytest.raises(ValueError):
        gp = GaussianProcess(x, y, nugget="a")

def test_GaussianProcess_n_params(x, y):
    "test the get_n_params method of GaussianProcess"

    gp = GaussianProcess(x, y)
    assert gp.n_params == x.shape[1] + 2

    gp = GaussianProcess(x, y, mean="x[0]")
    assert gp.n_params == 2 + x.shape[1] + 2

def test_GaussianProcess_nugget(x, y):
    "Tests the get_nugget method of GaussianProcess"

    gp = GaussianProcess(x, y)
    assert gp.nugget is None
    assert gp.nugget_type == "adaptive"

    gp.nugget = "fit"
    assert gp.nugget is None
    assert gp.nugget_type == "fit"

    gp.nugget = 1.
    assert_allclose(gp.nugget, 1.)
    assert gp.nugget_type == "fixed"

    gp.nugget = 0
    assert_allclose(gp.nugget, 0.)
    assert gp.nugget_type == "fixed"

    with pytest.raises(TypeError):
        gp.nugget = [1]

    with pytest.raises(ValueError):
        gp.nugget = "blah"

    with pytest.raises(ValueError):
        gp.nugget = -1.


@pytest.mark.parametrize("mean,nugget,sn", [(None, 0., 1.), (None, "adaptive", 0.),
                                            (MeanFunction("x[0]"), "fit", np.log(1.e-6))])
def test_GaussianProcess_theta(x, y, mean, nugget, sn):
    "test the theta property of GaussianProcess (effectively the same as fit)"

    # zero mean, zero nugget

    gp = GaussianProcess(x, y, mean=mean, nugget=nugget)

    theta = np.ones(gp.n_params)
    theta[-1] = sn

    gp.theta = theta

    switch = gp.mean.get_n_params(x)

    if nugget == "adaptive" or nugget == 0.:
        assert gp.nugget == 0.
        noise = 0.
    else:
        assert_allclose(gp.nugget, np.exp(sn))
        noise = np.exp(sn)*np.eye(x.shape[0])
    Q = gp.kernel.kernel_f(x, x, theta[switch:-1]) + noise
    ym = y - gp.mean.mean_f(x, theta[:switch])

    L_expect = np.linalg.cholesky(Q)
    invQt_expect = np.linalg.solve(Q, ym)
    logpost_expect = 0.5*(np.log(np.linalg.det(Q)) +
                          np.dot(ym, invQt_expect) +
                          gp.n*np.log(2.*np.pi))

    assert_allclose(L_expect, gp.L)
    assert_allclose(invQt_expect, gp.invQt)
    assert_allclose(logpost_expect, gp.current_logpost)

def test_GaussianProcess_priors(x, y):
    "test that priors are set properly"

    gp = GaussianProcess(x, y)

    assert gp.priors == [None, None, None, None, None]

    gp = GaussianProcess(x, y, priors=None)

    assert gp.priors == [None, None, None, None, None]

    priors = []
    gp = GaussianProcess(x, y, priors=priors)

    assert gp.priors == [None, None, None, None, None]

    priors = [None, None, None, None, None]
    gp = GausianProcess(x, y)

    assert gp.priors == priors

    priors = [None, NormalPrior(2., 2.), None, None, NormalPrior(3., 1.)]
    gp = GausianProcess(x, y, priors=priors)

    assert gp.priors[:-1] == priors[:-1]
    assert gp.priors[-1] is None

    priors = [None, NormalPrior(2., 2.), None, None]
    gp = GausianProcess(x, y, priors=priors)

    assert gp.priors[:-1] == priors
    assert gp.priors[-1] is None

    priors = [None, None, None]

    with pytest.raises(AssertionError):
        GausianProcess(x, y, priors=priors)

    with pytest.raises(TypeError):
        GaussianProcess(x, y, priors=1.)

    priors = [1., 2., 3., 4., 5.]

    with pytest.raises(TypeError):
        GaussianProcess(x, y, priors=priors)

@pytest.mark.parametrize("mean,nugget,sn", [(None, 0., 1.), (None, "adaptive", 0.),
                                            (MeanFunction("x[0]"), "fit", np.log(1.e-6))])
def test_GaussianProcess_fit_logposterior(x, y, mean, nugget, sn):
    "test the fit and logposterior methods of GaussianProcess"

    # zero mean, zero nugget

    gp = GaussianProcess(x, y, mean=mean, nugget=nugget)

    theta = np.ones(gp.n_params)
    theta[-1] = sn

    gp.fit(theta)

    switch = gp.mean.get_n_params(x)

    if nugget == "adaptive" or nugget == 0.:
        assert gp.nugget == 0.
        noise = 0.
    else:
        assert_allclose(gp.nugget, np.exp(sn))
        noise = np.exp(sn)*np.eye(x.shape[0])
    Q = gp.kernel.kernel_f(x, x, theta[switch:-1]) + noise
    ym = y - gp.mean.mean_f(x, theta[:switch])

    L_expect = np.linalg.cholesky(Q)
    invQt_expect = np.linalg.solve(Q, ym)
    logpost_expect = 0.5*(np.log(np.linalg.det(Q)) +
                          np.dot(ym, invQt_expect) +
                          gp.n*np.log(2.*np.pi))

    assert_allclose(L_expect, gp.L)
    assert_allclose(invQt_expect, gp.invQt)
    assert_allclose(logpost_expect, gp.current_logpost)
    assert_allclose(logpost_expect, gp.logposterior(theta))

def test_GaussianProcess_logposterior(x, y):
    "test logposterior method of GaussianProcess"

    # logposterior already tested, but check that parameters are re-fit if changed

    gp = GaussianProcess(x, y, nugget = 0.)

    theta = np.ones(gp.n_params)
    gp.fit(theta)

    theta = np.zeros(gp.n_params)

    Q = gp.kernel.kernel_f(x, x, theta[:-1])

    L_expect = np.linalg.cholesky(Q)
    invQt_expect = np.linalg.solve(Q, y)
    logpost_expect = 0.5*(np.log(np.linalg.det(Q)) +
                          np.dot(y, invQt_expect) +
                          gp.n*np.log(2.*np.pi))

    assert_allclose(logpost_expect, gp.logposterior(theta))
    assert_allclose(gp.L, L_expect)
    assert_allclose(invQt_expect, gp.invQt)
    assert_allclose(logpost_expect, gp.current_logpost)

@pytest.fixture
def dx():
    return 1.e-6

@pytest.mark.parametrize("mean,nugget,sn", [(None, 0., 1.), (None, "adaptive", 1.),
                                            ("x[0]", "fit", np.log(1.e-6))])
def test_GaussianProcess_logpost_deriv(x, y, dx, mean, nugget, sn):
    "test logposterior derivatives for GaussianProcess via finite differences"

    gp = GaussianProcess(x, y, mean=mean, nugget=nugget)

    n = gp.n_params
    theta = np.ones(n)
    theta[-1] = sn

    deriv = np.zeros(n)

    for i in range(n):
        dx_array = np.zeros(n)
        dx_array[i] = dx
        deriv[i] = (gp.logposterior(theta) - gp.logposterior(theta - dx_array))/dx

    assert_allclose(deriv, gp.logpost_deriv(theta), atol=1.e-7, rtol=1.e-5)

@pytest.mark.parametrize("mean,nugget,sn", [(None, 0., 1.), ("x[0]", "fit", np.log(1.e-6))])
def test_GaussianProcess_logpost_hessian(x, y, dx, mean, nugget, sn):
    "test the hessian method of GaussianProcess with finite differences"

    # zero mean, no nugget

    gp = GaussianProcess(x, y, nugget=0.)

    n = gp.n_params
    theta = np.ones(n)
    theta[-1] = sn

    hess = np.zeros((n, n))

    for i in range(n):
        for j in range(n):
            dx_array = np.zeros(n)
            dx_array[j] = dx
            hess[i, j] = (gp.logpost_deriv(theta)[i] - gp.logpost_deriv(theta - dx_array)[i])/dx

    assert_allclose(hess, gp.logpost_hessian(theta), rtol=1.e-5, atol=1.e-7)

@pytest.mark.parametrize("priors,nugget,sn", [([ NormalPrior(0.9, 0.5), None, NormalPrior(0.5, 2.),
                                              InvGammaPrior(2., 1.), None], 0., 0.),
                                           ([ None, NormalPrior(1.2, 0.2), None,
                                              GammaPrior(2., 1.), InvGammaPrior(2., 1.)], "fit", np.log(1.e-6))])
def test_GaussianProcess_priors(x, y, dx, priors, nugget, sn):
    "test that prior distributions are properly accounted for in posterior"

    gp = GaussianProcess(x, y, priors=priors, nugget=nugget)

    theta = np.ones(gp.n_params)
    theta[-1] = sn

    gp.fit(theta)

    if nugget == 0.:
        noise = 0.
    else:
        assert_allclose(gp.nugget, np.exp(sn))
        noise = np.exp(sn)*np.eye(x.shape[0])
    Q = gp.get_K_matrix() + noise

    L_expect = np.linalg.cholesky(Q)
    invQt_expect = np.linalg.solve(Q, y)
    logpost_expect = 0.5*(np.log(np.linalg.det(Q)) +
                          np.dot(y, invQt_expect) +
                          gp.n*np.log(2.*np.pi))

    for p, t in zip(priors, theta):
        if not p is None:
            logpost_expect -= p.logp(t)

    assert_allclose(L_expect, gp.L)
    assert_allclose(invQt_expect, gp.invQt)
    assert_allclose(logpost_expect, gp.current_logpost)
    assert_allclose(logpost_expect, gp.logposterior(theta))

    n = gp.n_params
    deriv = np.zeros(n)

    for i in range(n):
        dx_array = np.zeros(n)
        dx_array[i] = dx
        deriv[i] = (gp.logposterior(theta) - gp.logposterior(theta - dx_array))/dx

    assert_allclose(deriv, gp.logpost_deriv(theta), atol=1.e-5, rtol=1.e-4)

    hess = np.zeros((n, n))

    for i in range(n):
        for j in range(n):
            dx_array = np.zeros(n)
            dx_array[j] = dx
            hess[i, j] = (gp.logpost_deriv(theta)[i] - gp.logpost_deriv(theta - dx_array)[i])/dx

    assert_allclose(hess, gp.logpost_hessian(theta), rtol=1.e-5, atol=1.e-5)

def test_GaussianProcess_predict(x, y, dx):
    "test the predict method of GaussianProcess"

    # zero mean

    gp = GaussianProcess(x, y, nugget=0.)
    theta = np.ones(gp.n_params)

    gp.fit(theta)

    x_test = np.array([[2., 3., 4.]])

    mu, var, deriv = gp.predict(x_test)

    K = gp.kernel.kernel_f(x, x, theta[:-1])
    Ktest = gp.kernel.kernel_f(x_test, x, theta[:-1])

    mu_expect = np.dot(Ktest, gp.invQt)
    var_expect = np.exp(theta[-2]) - np.diag(np.dot(Ktest, np.linalg.solve(K, Ktest.T)))

    D = gp.D

    deriv_expect = np.zeros((1, D))

    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        deriv_expect[0, i] = (gp.predict(x_test)[0] - gp.predict(x_test - dx_array)[0])/dx

    assert_allclose(mu, mu_expect)
    assert_allclose(var, var_expect)
    assert_allclose(deriv, deriv_expect, atol=1.e-7, rtol=1.e-7)

    # check that reshaping works as expected

    x_test = np.array([2., 3., 4.])

    mu, var, deriv = gp.predict(x_test)

    assert_allclose(mu, mu_expect)
    assert_allclose(var, var_expect)
    assert_allclose(deriv, deriv_expect, atol=1.e-7, rtol=1.e-7)

    # check that with 1D input data can handle 1D prediction data correctly

    gp = GaussianProcess(y, y, nugget=0.)

    gp.fit(np.ones(gp.n_params))

    n_predict = 51
    mu, var, deriv = gp.predict(np.linspace(0., 1., n_predict))

    assert mu.shape == (n_predict,)
    assert var.shape == (n_predict,)
    assert deriv.shape == (n_predict, 1)

    # nonzero mean function

    gp = GaussianProcess(x, y, mean="x[0]", nugget=0.)

    theta = np.ones(gp.n_params)

    gp.fit(theta)

    x_test = np.array([[2., 3., 4.]])

    mu, var, deriv = gp.predict(x_test)

    switch = gp.mean.get_n_params(x)
    m = gp.mean.mean_f(x_test, theta[:switch])
    K = gp.kernel.kernel_f(x, x, theta[switch:-1])
    Ktest = gp.kernel.kernel_f(x_test, x, theta[switch:-1])

    mu_expect = m + np.dot(Ktest, gp.invQt)
    var_expect = np.exp(theta[-2]) - np.diag(np.dot(Ktest, np.linalg.solve(K, Ktest.T)))

    D = gp.D

    deriv_expect = np.zeros((1, D))

    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        deriv_expect[0, i] = (gp.predict(x_test)[0] - gp.predict(x_test - dx_array)[0])/dx

    assert_allclose(mu, mu_expect)
    assert_allclose(var, var_expect)
    assert_allclose(deriv, deriv_expect, atol=1.e-7, rtol=1.e-7)

    # check unc and deriv flags work

    _, var, deriv = gp.predict(x_test, unc=False, deriv=False)

    assert var is None
    assert deriv is None

    # check that the returned PredictResult works correctly

    pr = gp.predict(x_test)

    assert_allclose(pr.mean, mu_expect)
    assert_allclose(pr.unc, var_expect)
    assert_allclose(pr.deriv, deriv_expect, atol=1.e-7, rtol=1.e-7)

    assert_allclose(pr['mean'], mu_expect)
    assert_allclose(pr['unc'], var_expect)
    assert_allclose(pr['deriv'], deriv_expect, atol=1.e-7, rtol=1.e-7)

    assert_allclose(pr[0], mu_expect)
    assert_allclose(pr[1], var_expect)
    assert_allclose(pr[2], deriv_expect, atol=1.e-7, rtol=1.e-7)

    # check that calling gp is equivalent to predicting

    assert_allclose(gp(x_test), mu_expect)

def test_GaussianProcess_predict_nugget(x, y):
    "test that the nugget works correctly when making predictions"

    nugget = 1.e0

    gp = GaussianProcess(x, y, nugget=nugget)
    theta = np.ones(gp.n_params)

    gp.fit(theta)

    preds = gp.predict(x)

    K = gp.kernel.kernel_f(x, x, theta[:-1])

    var_expect = np.exp(theta[-2]) + nugget - np.diag(np.dot(K, np.linalg.solve(K + np.eye(gp.n)*nugget, K)))

    assert_allclose(preds.unc, var_expect, atol=1.e-7)

    preds = gp.predict(x, include_nugget=False)

    var_expect = np.exp(theta[-2]) - np.diag(np.dot(K, np.linalg.solve(K + np.eye(gp.n)*nugget, K)))

    assert_allclose(preds.unc, var_expect, atol=1.e-7)

def test_GaussianProcess_predict_variance():
    "confirm that caching factorized matrix produces stable variance predictions"

    x = np.linspace(0., 5., 21)
    y = x**2
    x = np.reshape(x, (-1, 1))
    nugget = 1.e-8

    gp = GaussianProcess(x, y, nugget=nugget)

    theta = np.array([-7.352408190715323, 15.041447753599755, 0.])
    gp.fit(theta)

    testing = np.reshape(np.linspace(0., 5., 101), (-1, 1))

    _, var, _ = gp.predict(testing)

    assert_allclose(np.zeros(101), var, atol = 1.e-3)

def test_GaussianProcess_predict_failures(x, y):
    "test situations where predict method of GaussianProcess should fail"

    gp = GaussianProcess(x, y)

    with pytest.raises(ValueError):
        gp.predict(np.array([2., 3., 4.]))

    theta = np.ones(gp.n_params)
    gp.fit(theta)

    with pytest.raises(AssertionError):
        gp.predict(np.ones((2, 2, 2)))

    with pytest.raises(AssertionError):
        gp.predict(np.array([[2., 4.]]))

def test_GaussianProcess_str(x, y):
    "Test function for string method"

    gp = GaussianProcess(x, y)
    assert (str(gp) == "Gaussian Process with {} training examples and {} input variables".format(x.shape[0], x.shape[1]))


import numpy as np
from mogp_emulator.MeanFunction import MeanFunction, MeanBase
from mogp_emulator.Kernel import Kernel, SquaredExponential, Matern52
from mogp_emulator.Priors import Prior
from scipy import linalg
from scipy.optimize import OptimizeResult
from mogp_emulator.linalg.cholesky import jit_cholesky

class GaussianProcess(object):
    """
    Implementation of a Gaussian Process Emulator.

    This class provides a representation of a Gaussian Process Emulator. It contains
    methods for fitting the GP to a given set of hyperparameters, computing the
    negative log marginal likelihood plus prior (so negative log posterior) and its
    derivatives, and making predictions on unseen data. Note that routines to
    estimate hyperparameters are not included in the class definition, and are instead
    provided externally to facilitate implementation of high performance versions.

    The required arguments to initialize a GP is a set of training data consisting
    of the inputs and targets for each of those inputs. These must be numpy arrays
    whose first axis is of the same length. Targets must be a 1D array, while inputs
    can be 2D or 1D. If inputs is 1D, then it is assumed that the length of the second
    axis is unity.

    Optional arguments are the particular mean function to use (default is zero mean),
    the covariance kernel to use (default is the squared exponential covariance),
    a list of prior distributions for each hyperparameter (default is no prior information
    on any hyperparameters) and the method for handling the nugget parameter.
    The nugget is additional "noise" that is added to the diagonal of the covariance
    kernel as a variance. This nugget can represent uncertainty in the target values
    themselves, or simply be used to stabilize the numerical inversion of the covariance
    matrix. The nugget can be fixed (a non-negative float), can be found adaptively
    (by passing the string ``"adaptive"`` to make the noise only as large as necessary
    to successfully invert the covariance matrix), or can be fit as a hyperparameter
    (by passing the string ``"fit"``).

    The internal emulator structure involves arrays for the inputs, targets, and hyperparameters.
    Other useful information are the number of training examples ``n``, the number of input
    parameters ``D``, and the number of hyperparameters ``n_params``. These parameters can
    be obtained externally by accessing these attributes

    Example: ::

        >>> import numpy as np
        >>> from mogp_emulator import GaussianProcess
        >>> x = np.array([[1., 2., 3.], [4., 5., 6.]])
        >>> y = np.array([4., 6.])
        >>> gp = GaussianProcess(x, y)
        >>> print(gp)
        Gaussian Process with 2 training examples and 3 input variables
        >>> gp.n
        2
        >>> gp.D
        3
        >>> gp.n_params
        5
        >>> gp.fit(np.zeros(gp.n_params))
        >>> x_predict = np.array([[2., 3., 4.], [7., 8., 9.]])
        >>> gp.predict(x_predict)
        (array([4.74687618, 6.84934016]), array([0.01639298, 1.05374973]),
        array([[8.91363045e-05, 7.18827798e-01, 3.74439445e-16],
               [4.64005897e-06, 3.74191346e-02, 1.94917337e-17]]))

    """
    def __init__(self, inputs, targets, mean=None, kernel=SquaredExponential(), priors=None,
                 nugget="adaptive", inputdict = {}, use_patsy=True):
        """
        Create a new GaussianProcess Emulator

        Creates a new GaussianProcess Emulator from either the input data and targets to be fit and
        optionally a mean function, covariance kernel, and nugget parameter/method.

        Required arguments are numpy arrays ``inputs`` and ``targets``, described below.
        Additional arguments are ``mean`` to specify the mean function (default is
        ``None`` for zero mean), ``kernel`` to specify the covariance kernel (default
        is the squared exponential kernel), ``priors`` to indicate prior distributions
        on the hyperparameters, and ``nugget`` to specify how the nugget
        parameter is handled (see below; default is to fit the nugget adaptively).

        ``inputs`` is a 2D array-like object holding the input data, whose shape is
        ``n`` by ``D``, where ``n`` is the number of training examples to be fit and ``D``
        is the number of input variables to each simulation. If ``inputs`` is a 1D array,
        it is assumed that ``D = 1``.

        ``targets`` is the target data to be fit by the emulator, also held in an array-like
        object. This must be a 1D array of length ``n``.

        ``prior`` must be a list of length ``n_params`` whose elements are either ``Prior``-derived
        objects or ``None``.  Each element is used as the prior for the corresponding parameter (with
        ``None`` indicating an uninformative prior).  Passing the empty list or ``None`` as this
        argument (in its entirety) may be used as an abbreviation for a list of ``n_params`` where
        all list elements are ``None``.

        ``nugget`` controls how additional noise is added to the emulator targets when fitting.
        This can be specified in several ways. If a string is provided, it can take the
        values of ``"adaptive"`` or ``"fit"``, which indicate that the nugget will be
        chosen in the fitting process. ``"adaptive"`` means that the nugget will be made only
        as large as necessary to invert the covariance matrix, while ``"fit"`` means that
        the nugget will be treated as a hyperparameter to be optimized. Alternatively,
        a non-negative float can be used to specify a fixed noise level. If no value is
        specified for the ``nugget`` parameter, ``"adaptive"`` is the default.


        :param inputs: Numpy array holding emulator input parameters. Must be 1D with length
                       ``n`` or 2D with shape ``n`` by ``D``, where ``n`` is the number of
                       training examples and ``D`` is the number of input parameters for
                       each output.
        :type inputs: ndarray
        :param targets: Numpy array holding emulator targets. Must be 1D with length ``n``
        :type targets: ndarray
        :param mean: Mean function to be used (optional, default is ``None`` for a zero mean)
        :type mean: None or MeanFunction
        :param kernel: Covariance kernel to be used (optional, default is Squared Exponential)
                       Can provide either a ``Kernel`` object or a string matching the
                       kernel type to be used.
        :type kernel: Kernel or str
        :param priors: List of priors to be used. Must be None (default) or an empty list
                       (indicates uninformative priors) or list of length ``n_params``.
                       Any parameter for which you wish to specify an uninformative prior,
                       pass ``None``. Number of parameters is the number of parameters in
                       the mean function plus ``D + 2`` (one correlation length per input
                       plus a covariance scale and a nugget). If the nugget will not be fit,
                       the list can have length ``n_params - 1``.
        :type priors: list or None
        :param nugget: Noise to be added to the diagonal, specified as a string or a float.
                       A non-negative float specifies the noise level explicitly, while a string
                       indicates that the nugget will be found via fitting, either as ``"adaptive"``
                       or ``"fit"`` (see above for a description). Default is ``"adaptive"``.
        :type nugget: float or str
        :returns: New ``GaussianProcess`` instance
        :rtype: GaussianProcess

        """
        inputs = np.array(inputs)
        if inputs.ndim == 1:
            inputs = np.reshape(inputs, (-1, 1))
        assert inputs.ndim == 2

        targets = np.array(targets)
        assert targets.ndim == 1
        assert targets.shape[0] == inputs.shape[0]

        if not issubclass(type(mean), MeanBase):
            if not (mean is None or isinstance(mean, str)):
                raise ValueError("provided mean function must be a subclass of MeanFunction,"+
                                 " a string formula, or None")

        if isinstance(kernel, str):
            if kernel == "SquaredExponential":
                kernel = SquaredExponential()
            elif kernel == "Matern52":
                kernel = Matern52()
            else:
                raise ValueError("provided kernel '{}' not a supported kernel type".format(kernel))
        if not issubclass(type(kernel), Kernel):
            raise ValueError("provided kernel is not a subclass of Kernel")

        self._inputs = inputs
        self._targets = targets

        if not issubclass(type(mean), MeanBase):
            self.mean = MeanFunction(mean, inputdict, use_patsy)
        else:
            self.mean = mean

        self.kernel = kernel

        self.nugget = nugget

        self.priors = priors

        self._theta = None

    @property
    def inputs(self):
        """
        Returns inputs for the emulator as a numpy array

        :returns: Emulator inputs, 2D array with shape ``(n, D)``
        :rtype: ndarray
        """
        return self._inputs

    @property
    def targets(self):
        """
        Returns targets for the emulator as a numpy array

        :returns: Emulator targets, 1D array with shape ``(n,)``
        :rtype: ndarray
        """
        return self._targets

    @property
    def n(self):
        """
        Returns number of training examples for the emulator

        :returns: Number of training examples for the emulator object
        :rtype: int
        """

        return self.inputs.shape[0]

    @property
    def D(self):
        """
        Returns number of inputs for the emulator

        :returns: Number of inputs for the emulator object
        :rtype: int
        """

        return self.inputs.shape[1]

    @property
    def n_params(self):
        """
        Returns number of hyperparameters

        Returns the number of hyperparameters for the emulator. The number depends on the
        choice of mean function, covariance function, and nugget strategy, and possibly the
        number of inputs for certain choices of the mean function.

        :returns: Number of hyperparameters
        :rtype: int
        """

        return self.mean.get_n_params(self.inputs) + self.D + 2

    @property
    def nugget_type(self):
        """
        Returns method used to select nugget parameter

        Returns a string indicating how the nugget parameter is treated, either ``"adaptive"``,
        ``"fit"``, or ``"fixed"``. This is automatically set when changing the ``nugget``
        property.

        :returns: Current nugget fitting method
        :rtype: str
        """
        return self._nugget_type

    @property
    def nugget(self):
        """
        Returns emulator nugget parameter

        Returns current value of the nugget parameter. If the nugget is to be selected
        adaptively or by fitting the emulator and the nugget has not been fit,
        returns ``None``.

        :returns: Current nugget value, either a float or ``None``
        :rtype: float or None

        The ``nugget`` parameter controls how noise is added to the covariance matrix in order to
        stabilize the inversion or smooth the emulator predictions. If ``nugget`` is a non-negative
        float, then that particular value is used for the nugget. Note that setting this parameter
        to be zero enforces that the emulator strictly interpolates between points. Alternatively,
        a string can be provided. A value of ``"fit"`` means that the nugget is treated as a
        hyperparameter, and is the last entry in the ``theta`` array. Alternatively, if ``nugget``
        is set to be ``"adaptive"``, the fitting routine will adaptively make the noise parameter
        as large as is needed to ensure that the emulator can be fit.

        Internally, this modifies both the way the nugget is chosen (which can be determined
        via the ``nugget_type`` property) and the value itself (the ``nugget`` property)

        :param nugget: Noise to be added to the diagonal, specified as a string or a float.
                       A non-negative float specifies the noise level explicitly, while a string
                       indicates that the nugget will be found via fitting, either as ``"adaptive"``
                       or ``"fit"`` (see above for a description).
        :type nugget: float or str
        :returns: None
        :rtype: None
        """

        return self._nugget

    @nugget.setter
    def nugget(self, nugget):
        """
        Set the nugget parameter for the emulator

        Method for changing the ``nugget`` parameter for the emulator. When a new emulator is
        initilized, this is set to None.

        The ``nugget`` parameter controls how noise is added to the covariance matrix in order to
        stabilize the inversion or smooth the emulator predictions. If ``nugget`` is a non-negative
        float, then that particular value is used for the nugget. Note that setting this parameter
        to be zero enforces that the emulator strictly interpolates between points. Alternatively,
        a string can be provided. A value of ``"fit"`` means that the nugget is treated as a
        hyperparameter, and is the last entry in the ``theta`` array. Alternatively, if ``nugget``
        is set to be ``"adaptive"``, the fitting routine will adaptively make the noise parameter
        as large as is needed to ensure that the emulator can be fit.

        Internally, this modifies both the way the nugget is chosen (which can be determined
        via the ``nugget_type`` property) and the value itself (the ``nugget`` property)

        :param nugget: Noise to be added to the diagonal, specified as a string or a float.
                       A non-negative float specifies the noise level explicitly, while a string
                       indicates that the nugget will be found via fitting, either as ``"adaptive"``
                       or ``"fit"`` (see above for a description).
        :type nugget: float or str
        :returns: None
        :rtype: None
        """

        if not isinstance(nugget, (str, float)):
            try:
                nugget = float(nugget)
            except TypeError:
                raise TypeError("nugget parameter must be a string or a non-negative float")

        if isinstance(nugget, str):
            if nugget == "adaptive":
                self._nugget_type = "adaptive"
            elif nugget == "fit":
                self._nugget_type = "fit"
            else:
                raise ValueError("bad value of nugget, must be a float or 'adaptive' or 'fit'")
            self._nugget = None
        else:
            if nugget < 0.:
                raise ValueError("nugget parameter must be non-negative")
            self._nugget_type = "fixed"
            self._nugget = float(nugget)

    @property
    def theta(self):
        """
        Returns emulator hyperparameters

        Returns current hyperparameters for the emulator as a numpy array if they have been fit.
        If no parameters have been fit, returns ``None``. Note that the number of parameters
        depends on the mean function, so the length of this array will vary across instances.

        :returns: Current parameter values (numpy array of length ``n_params``), or ``None`` if the
                  parameters have not been fit.
        :rtype: ndarray or None

        When set, pre-calculates the matrices needed to compute the log-likelihood and its derivatives
        and make subsequent predictions. This is called any time the hyperparameter values are
        changed in order to ensure that all the information is needed to evaluate the
        log-likelihood and its derivatives, which are needed when fitting the optimal
        hyperparameters.

        The method computes the mean function and covariance matrix and inverts the covariance
        matrix using the method specified by the value of ``nugget``. The factorized matrix
        and the product of the inverse with the difference between the targets and the mean
        are cached for later use, and the negative marginal log-likelihood is also cached.
        This method has no return value, but it does modify the state of the object.

        :param theta: Values of the hyperparameters to use in fitting. Must be a numpy
                      array with length ``n_params``
        :type theta: ndarray
        """

        return self._theta

    @theta.setter
    def theta(self, theta):
        """
        Fits the emulator and sets the parameters (property-based setter alias for ``fit``)

        Pre-calculates the matrices needed to compute the log-likelihood and its derivatives
        and make subsequent predictions. This is called any time the hyperparameter values are
        changed in order to ensure that all the information is needed to evaluate the
        log-likelihood and its derivatives, which are needed when fitting the optimal
        hyperparameters.

        The method computes the mean function and covariance matrix and inverts the covariance
        matrix using the method specified by the value of ``nugget``. The factorized matrix
        and the product of the inverse with the difference between the targets and the mean
        are cached for later use, and the negative marginal log-likelihood is also cached.
        This method has no return value, but it does modify the state of the object.

        :param theta: Values of the hyperparameters to use in fitting. Must be a numpy
                      array with length ``n_params``
        :type theta: ndarray
        :returns: None
        """
        self.fit(theta)

    @property
    def priors(self):
        """
        The current list priors used in computing the log posterior

        To set the priors, must be a list or ``None``. Entries can be ``None`` or a subclass of ``Prior``.
        ``None`` indicates weak prior information. An empty list or ``None`` means all uninformative
        priors. Otherwise list should have the same length as the number of hyperparameters,
        or alternatively can be one shorter than the number of hyperparameters
        if ``nugget_type`` is ``"adaptive"`` or ``"fixed"`` meaning that the nugget hyperparameter
        is not fit but is instead fixed or found adaptively. If the nugget hyperparameter is not fit,
        the prior for the nugget will automatically be set to ``None`` even if a distribution is
        provided.
        """
        return self._priors

    @priors.setter
    def priors(self, priors):
        """
        Sets the priors to a list of prior objects/None

        Sets the priors, must be a list or ``None``. Entries can be ``None`` or a subclass of ``Prior``.
        ``None`` indicates weak prior information. An empty list or ``None`` means all uninformative
        priors. Otherwise list should have the same length as the number of hyperparameters,
        or alternatively can be one shorter than the number of hyperparameters
        if ``nugget_type`` is ``"adaptive"`` or ``"fixed"`` meaning that the nugget hyperparameter
        is not fit but is instead fixed or found adaptively. If the nugget hyperparameter is not fit,
        the prior for the nugget will automatically be set to ``None`` even if a distribution is
        provided.
        """

        if priors is None:
            priors = []

        if not isinstance(priors, list):
            raise TypeError("priors must be a list of Prior-derived objects")

        if len(priors) == 0:
            priors = self.n_params*[None]

        if self.nugget_type in ["adaptive", "fixed"]:
            if len(priors) == self.n_params - 1:
                priors.append(None)

        if not len(priors) == self.n_params:
            raise ValueError("bad length for priors; must have length n_params")

        if self.nugget_type in ["adaptive", "fixed"]:
            if not priors[-1] is None:
                priors[-1] = None

        for p in priors:
            if not p is None and not issubclass(type(p), Prior):
                raise TypeError("priors must be a list of Prior-derived objects")

        self._priors = list(priors)


    def get_K_matrix(self):
        """
        Returns current value of the covariance matrix as a numpy array. Does not include the nugget
        parameter, as this is dependent on how the nugget is fit.
        """
        switch = self.mean.get_n_params(self.inputs)

        return self.kernel.kernel_f(self.inputs, self.inputs, self.theta[switch:-1])

    def fit(self, theta):
        """
        Fits the emulator and sets the parameters

        Pre-calculates the matrices needed to compute the log-likelihood and its derivatives
        and make subsequent predictions. This is called any time the hyperparameter values are
        changed in order to ensure that all the information is needed to evaluate the
        log-likelihood and its derivatives, which are needed when fitting the optimal
        hyperparameters.

        The method computes the mean function and covariance matrix and inverts the covariance
        matrix using the method specified by the value of ``nugget_type``. The factorized matrix
        and the product of the inverse with the difference between the targets and the mean
        are cached for later use, and the negative marginal log-likelihood is also cached.
        This method has no return value, but it does modify the state of the object.

        :param theta: Values of the hyperparameters to use in fitting. Must be a numpy
                      array with length ``n_params``
        :type theta: ndarray
        :returns: None
        """

        theta = np.array(theta)

        assert theta.shape == (self.n_params,), "bad shape for hyperparameters"

        self._theta = theta

        switch = self.mean.get_n_params(self.inputs)

        m = self.mean.mean_f(self.inputs, self.theta[:switch])
        Q = self.kernel.kernel_f(self.inputs, self.inputs, self.theta[switch:-1])

        if self.nugget_type == "adaptive":
            self.L, self._nugget = jit_cholesky(Q)
        else:
            if self.nugget_type == "fit":
                self._nugget = np.exp(self.theta[-1])
            Q += self._nugget*np.eye(self.n)
            self.L = linalg.cholesky(Q, lower=True)

        self.invQt = linalg.cho_solve((self.L, True), self.targets - m)

        self.current_logpost = 0.5*(2.0*np.sum(np.log(np.diag(self.L))) +
                                    np.dot(self.targets - m, self.invQt) +
                                    self.n*np.log(2. * np.pi))

        for i in range(self.n_params):
            if not self._priors[i] is None:
                self.current_logpost -= self._priors[i].logp(self.theta[i])


    def logposterior(self, theta):
        """
        Calculate the negative log-posterior at a particular value of the hyperparameters

        Calculate the negative log-posterior for the given set of parameters. Calling this
        method sets the parameter values and computes the needed inverse matrices in order
        to evaluate the log-posterior and its derivatives. In addition to returning the
        log-posterior value, it stores the current value of the hyperparameters and
        log-posterior in attributes of the object.

        :param theta: Value of the hyperparameters. Must be array-like with shape ``(n_params,)``
        :type theta: ndarray
        :returns: negative log-posterior
        :rtype: float
        """

        if self.theta is None or not np.allclose(theta, self.theta, rtol=1.e-10, atol=1.e-15):
            self.fit(theta)

        return self.current_logpost

    def logpost_deriv(self, theta):
        """
        Calculate the partial derivatives of the negative log-posterior

        Calculate the partial derivatives of the negative log-posterior with respect to
        the hyperparameters. Note that this function is normally used only when fitting
        the hyperparameters, and it is not needed to make predictions.

        During normal use, the ``loglike_deriv`` method is called after evaluating the
        ``logposterior`` method. The implementation takes advantage of this by reusing
        cached results, as the factorized covariance matrix is expensive to compute and is
        used by the ``logposterior``, ``logpost_deriv``, and ``logpost_hessian`` methods.
        If the function is evaluated with a different set of parameters than was previously
        used to set the log-posterior, the method calls ``fit`` (and subsequently resets
        the cached information).

        :param theta: Value of the hyperparameters. Must be array-like with shape
                      ``(n_params,)``
        :type theta: ndarray
        :returns: partial derivatives of the negative log-posterior with respect to the
                  hyperparameters (array with shape ``(n_params,)``)
        :rtype: ndarray
        """

        theta = np.array(theta)

        assert theta.shape == (self.n_params,), "bad shape for new parameters"

        if self.theta is None or not np.allclose(theta, self.theta, rtol=1.e-10, atol=1.e-15):
            self.fit(theta)

        partials = np.zeros(self.n_params)

        switch = self.mean.get_n_params(self.inputs)

        dmdtheta = self.mean.mean_deriv(self.inputs, self.theta[:switch])
        dKdtheta = self.kernel.kernel_deriv(self.inputs, self.inputs, self.theta[switch:-1])

        partials[:switch] = -np.dot(dmdtheta, self.invQt)

        for d in range(self.D + 1):
            invQ_dot_dKdtheta_trace = np.trace(linalg.cho_solve((self.L, True), dKdtheta[d]))
            partials[switch + d] = -0.5*(np.dot(self.invQt, np.dot(dKdtheta[d], self.invQt)) -
                                         invQ_dot_dKdtheta_trace)

        if self.nugget_type == "fit":
            nugget = np.exp(self.theta[-1])
            partials[-1] = 0.5*nugget*(np.trace(linalg.cho_solve((self.L, True), np.eye(self.n))) -
                                       np.dot(self.invQt, self.invQt))

        for i in range(self.n_params):
            if not self._priors[i] is None:
                partials[i] -= self._priors[i].dlogpdtheta(self.theta[i])

        return partials

    def logpost_hessian(self, theta):
        """
        Calculate the Hessian of the negative log-posterior

        Calculate the Hessian of the negative log-posterior with respect to
        the hyperparameters. Note that this function is normally used only when fitting
        the hyperparameters, and it is not needed to make predictions. It is also used
        to estimate an appropriate step size when fitting hyperparameters using
        the lognormal approximation or MCMC sampling.

        When used in an optimization routine, the ``logpost_hessian`` method is called after
        evaluating the ``logposterior`` method. The implementation takes advantage of
        this by storing the inverse of the covariance matrix, which is expensive to
        compute and is used by the ``logposterior`` and ``logpost_deriv`` methods as well.
        If the function is evaluated with a different set of parameters than was previously
        used to set the log-posterior, the method calls ``fit`` to compute the needed
        information and changes the cached values.

        :param theta: Value of the hyperparameters. Must be array-like with shape
                      ``(n_params,)``
        :type theta: ndarray
        :returns: Hessian of the negative log-posterior (array with shape
                  ``(n_params, n_params)``)
        :rtype: ndarray
        """

        assert theta.shape == (self.n_params,), "Parameter vector must have length number of inputs + 1"

        if self.theta is None or not np.allclose(theta, self.theta, rtol=1.e-10, atol=1.e-15):
            self.fit(theta)

        hessian = np.zeros((self.n_params, self.n_params))

        switch = self.mean.get_n_params(self.inputs)

        dmdtheta = self.mean.mean_deriv(self.inputs, self.theta[:switch])
        d2mdtheta2 = self.mean.mean_hessian(self.inputs, self.theta[:switch])
        dKdtheta = self.kernel.kernel_deriv(self.inputs, self.inputs, self.theta[switch:-1])
        d2Kdtheta2 = self.kernel.kernel_hessian(self.inputs, self.inputs, self.theta[switch:-1])

        hessian[:switch, :switch] = -(np.dot(d2mdtheta2, self.invQt) -
                                      np.dot(dmdtheta, linalg.cho_solve((self.L, True),
                                                                        np.transpose(dmdtheta))))

        hessian[:switch, switch:-1] = np.dot(dmdtheta,
                                             linalg.cho_solve((self.L, True),
                                                              np.transpose(np.dot(dKdtheta, self.invQt))))

        hessian[switch:-1, :switch] = np.transpose(hessian[:switch, switch:-1])

        for d1 in range(self.D + 1):
            invQ_dot_d1 = linalg.cho_solve((self.L, True), dKdtheta[d1])
            for d2 in range(self.D + 1):
                invQ_dot_d2 = linalg.cho_solve((self.L, True), dKdtheta[d2])
                invQ_dot_d1d2 = linalg.cho_solve((self.L, True), d2Kdtheta2[d1, d2])
                term_1 = np.linalg.multi_dot([self.invQt,
                                              2.*np.dot(dKdtheta[d1], invQ_dot_d2) - d2Kdtheta2[d1, d2],
                                              self.invQt])
                term_2 = np.trace(np.dot(invQ_dot_d1, invQ_dot_d2) - invQ_dot_d1d2)
                hessian[switch + d1, switch + d2] = 0.5*(term_1 - term_2)

        if self.nugget_type == "fit":
            nugget = np.exp(self.theta[-1])
            invQinvQt = linalg.cho_solve((self.L, True), self.invQt)
            hessian[:switch, -1] = nugget*np.dot(dmdtheta, invQinvQt)
            for d in range(self.D + 1):
                hessian[switch + d, -1] = nugget*(np.linalg.multi_dot([self.invQt, dKdtheta[d], invQinvQt]) -
                                                  0.5*np.trace(linalg.cho_solve((self.L, True),
                                                                                np.dot(dKdtheta[d],
                                                                                       linalg.cho_solve((self.L, True),
                                                                                                        np.eye(self.n))))))

            hessian[-1, -1] = 0.5*nugget*(np.trace(linalg.cho_solve((self.L, True), np.eye(self.n))) -
                                                   np.dot(self.invQt, self.invQt))
            hessian[-1, -1] += nugget**2*(np.dot(self.invQt, invQinvQt) -
                                          0.5*np.trace(linalg.cho_solve((self.L, True),
                                                                        linalg.cho_solve((self.L, True),
                                                                                         np.eye(self.n)))))

            hessian[-1, :-1] = np.transpose(hessian[:-1, -1])

        for i in range(self.n_params):
            if not self._priors[i] is None:
                hessian[i, i] -= self._priors[i].d2logpdtheta2(self.theta[i])

        return hessian

    def predict(self, testing, unc=True, deriv=True, include_nugget=True):
        """
        Make a prediction for a set of input vectors for a single set of hyperparameters

        Makes predictions for the emulator on a given set of input vectors. The input vectors
        must be passed as a ``(n_predict, D)``, ``(n_predict,)`` or ``(D,)`` shaped array-like
        object, where ``n_predict`` is the number of different prediction points under
        consideration and ``D`` is the number of inputs to the emulator. If the prediction
        inputs array is 1D and ``D == 1`` for the GP instance, then the 1D array must have
        shape ``(n_predict,)``. Otherwise, if the array is 1D it must have shape
        ``(D,)``, and the method assumes ``n_predict == 1``. The prediction is returned as an
        ``(n_predict, )`` shaped numpy array as the first return value from the method.

        Optionally, the emulator can also calculate the variances in the predictions
        and the derivatives with respect to each input parameter. If the uncertainties are
        computed, they are returned as the second output from the method as an ``(n_predict,)``
        shaped numpy array. If the derivatives are computed, they are returned as the third
        output from the method as an ``(n_predict, D)`` shaped numpy array.

        The final input to the method determines if the predictive variance should include
        the nugget or not. For situations where the nugget represents observational error
        and predictions are estimating the true underlying function, this should be set to
        ``False``. However, most other cases should probably include the nugget, as the
        emulator is using it to represent some of the uncertainty in the underlying simulator,
        so the default value is ``True``.

        :param testing: Array-like object holding the points where predictions will be made.
                        Must have shape ``(n_predict, D)`` or ``(D,)`` (for a single prediction)
        :type testing: ndarray
        :param unc: (optional) Flag indicating if the uncertainties are to be computed.
                    If ``False`` the method returns ``None`` in place of the uncertainty
                    array. Default value is ``True``.
        :type unc: bool
        :param deriv: (optional) Flag indicating if the derivatives are to be computed.
                      If ``False`` the method returns ``None`` in place of the derivative
                      array. Default value is ``True``.
        :type deriv: bool
        :param include_nugget: (optional) Flag indicating if the nugget should be included
                               in the predictive variance. Only relevant if ``unc = True``.
                               Default is ``True``.
        :type include_nugget: bool
        :returns: Tuple of numpy arrays holding the predictions, uncertainties, and derivatives,
                  respectively. Predictions and uncertainties have shape ``(n_predict,)``
                  while the derivatives have shape ``(n_predict, D)``. If the ``unc`` or
                  ``deriv`` flags are set to ``False``, then those arrays are replaced by
                  ``None``.
        :rtype: tuple
        """

        if self.theta is None:
            raise ValueError("hyperparameters have not been fit for this Gaussian Process")

        testing = np.array(testing)
        if self.D == 1 and testing.ndim == 1:
            testing = np.reshape(testing, (-1, 1))
        elif testing.ndim == 1:
            testing = np.reshape(testing, (1, len(testing)))
        assert testing.ndim == 2

        n_testing, D = np.shape(testing)
        assert D == self.D

        switch = self.mean.get_n_params(testing)
        mtest = self.mean.mean_f(testing, self.theta[:switch])
        Ktest = self.kernel.kernel_f(self.inputs, testing, self.theta[switch:-1])

        mu = mtest + np.dot(Ktest.T, self.invQt)

        var = None
        if unc:
            sigma_2 = np.exp(self.theta[-2])

            if include_nugget:
                sigma_2 += self._nugget

            var = np.maximum(sigma_2 - np.sum(Ktest*linalg.cho_solve((self.L, True), Ktest), axis=0),
                             0.)

        inputderiv = None
        if deriv:
            inputderiv = np.zeros((n_testing, self.D))
            mean_deriv = self.mean.mean_inputderiv(testing, self.theta[:switch])
            kern_deriv = self.kernel.kernel_inputderiv(testing, self.inputs, self.theta[switch:-1])
            inputderiv = np.transpose(mean_deriv + np.dot(kern_deriv, self.invQt))

        return PredictResult(mean=mu, unc=var, deriv=inputderiv)

    def __call__(self, testing):
        """A Gaussian process object is callable: calling it is the same as
        calling `predict` without uncertainty and derivative
        predictions, and extracting the zeroth component for the
        'mean' prediction.
        """
        return (self.predict(testing, unc=False, deriv=False)[0])

    def __str__(self):
        """
        Returns a string representation of the model

        :returns: A string representation of the model (indicates number of training examples
                  and inputs)
        :rtype: str
        """

        return ("Gaussian Process with " + str(self.n) + " training examples and " +
                str(self.D) + " input variables")

class PredictResult(dict):
    """
    Prediction results object

    Dictionary-like object containing mean, uncertainty (variance), and derivatives with
    respect to the inputs of an emulator prediction. Values can be accessed like a dictionary
    with keys ``'mean'``, ``'unc'``, and ``'deriv'`` (or indices 0, 1, and 2 for the mean,
    uncertainty, and derivative for backwards compatability), or using attributes
    (``p.mean`` if ``p`` is an instance of ``PredictResult``). Also supports iteration and
    unpacking with the ordering ``(mean, unc, deriv)`` to be consistent with indexing behavior.

    Code is mostly based on scipy's ``OptimizeResult`` class, with some additional code to
    support iteration and integer indexing.

    :ivar mean: Predicted mean for each input point. Numpy array with shape ``(n_predict,)``
    :type mean: ndarray
    :ivar unc: Predicted variance for each input point. Numpy array with shape ``(n_predict,)``
    :type mean: ndarray
    :ivar deriv: Predicted derivative with respect to the inputs for each input point.
                 Numpy array with shape ``(n_predict, D)``
    :type deriv: ndarray
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __iter__(self):
        self.index = 0
        return self

    def __next__(self):
        if self.index == 0:
            self.index += 1
            return self['mean']
        elif self.index == 1:
            self.index += 1
            return self['unc']
        elif self.index == 2:
            self.index += 1
            return self['deriv']
        else:
            raise StopIteration


    def __getitem__(self, key):
        if not isinstance(key, (int, str)):
            raise KeyError(key)
        if key == 0:
            newkey = "mean"
        elif key == 1:
            newkey = "unc"
        elif key == 2:
            newkey = "deriv"
        else:
            newkey = key

        return dict.__getitem__(self, newkey)

    def __repr__(self):
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in zip(["mean", "unc", "deriv"], self)])
        else:
            return self.__class__.__name__ + "()"


# -*- coding: utf-8 -*-
"""
Created on Wed Apr 24 05:32:45 2013

@author: tisimst
"""
import random
from copy import copy
import numpy as np
import scipy.stats as ss

def k_fold_cross_validation(X, K, randomise=False):
    """
    Generates K (training, validation) pairs from the items in X.

    Each pair is a partition of X, where validation is an iterable
    of length len(X)/K. So each training iterable is of length (K-1)*len(X)/K.

    If randomise is true, a copy of X is shuffled before partitioning,
    otherwise its order is preserved in training and validation.
    ## {{{ http://code.activestate.com/recipes/521906/ (r3)
    """
    if randomise:
        X = list(X)
        random.shuffle(X)
    for k in range(K):
        training = [x for i, x in enumerate(X) if i % K != k]
        validation = [x for i, x in enumerate(X) if i % K == k]
        yield training, validation

        
def integer_bisect(bound, f):
    """
    Finds a pair of integers (a,b) such that f(a) <= 0 < f(b) and |a - b| == 1.
    On entry, assumes that f(bound[0]) <= 0 < f(bound[1])
    """
    if bound[1] - bound[0] == 1:
        return bound
    else:
        midpoint = round(bound[0] + bound[1] / 2.0)
        if f(midpoint) <= 0:
            return integer_bisect((midpoint, bound[1]), f)
        else:
            return integer_bisect((bound[0], midpoint), f)
